import os

import numpy as np
import openmdao.api as om
from scipy.interpolate import PchipInterpolator


class PoseOptimization(object):
    def __init__(self, wt_init, modeling_options, analysis_options):
        self.wt_init = wt_init
        self.modeling = modeling_options
        self.opt = analysis_options

        self.nlopt_methods = [
            "GN_DIRECT",
            "GN_DIRECT_L",
            "GN_DIRECT_L_NOSCAL",
            "GN_ORIG_DIRECT",
            "GN_ORIG_DIRECT_L",
            "GN_AGS",
            "GN_ISRES",
            "LN_COBYLA",
            "LD_MMA",
            "LD_CCSAQ",
            "LD_SLSQP",
        ]

        self.scipy_methods = [
            "SLSQP",
            "Nelder-Mead",
            "COBYLA",
        ]

        self.pyoptsparse_methods = [
            "SNOPT",
            "CONMIN",
            "NSGA2",
        ]

        self.floating_solve_component = 'floatingse'
        self.floating_period_solve_component = 'floatingse'
        
    def _get_step_size(self):
        # If a step size for the driver-level finite differencing is provided, use that step size. Otherwise use a default value.
        return (
            1.0e-6
            if not "step_size" in self.opt["driver"]["optimization"]
            else self.opt["driver"]["optimization"]["step_size"]
        )

    def _set_optimizer_properties(self, wt_opt, options_keys=[], opt_settings_keys=[], mapped_keys={}):
        """
        Set the optimizer properties, both the `driver.options` and
        `driver.opt_settings`. See OpenMDAO documentation on drivers
        to determine which settings are set by either options or
        opt_settings.

        Parameters
        ----------
        wt_opt : OpenMDAO problem object
            The wind turbine problem object.
        options_keys : list
            List of keys for driver options to be set.
        opt_settings_keys: list
            List of keys for driver opt_settings to be set.
        mapped_keys: dict
            Key pairs where the yaml name differs from what's expected
            by the driver. Specifically, the key is what's given in the yaml
            and the value is what's expected by the driver.

        Returns
        -------
        wt_opt : OpenMDAO problem object
            The updated wind turbine problem object with driver settings applied.
        """

        opt_options = self.opt["driver"]["optimization"]

        # Loop through all of the options provided and set them in the OM driver object
        for key in options_keys:
            if key in opt_options:
                if key in mapped_keys:
                    wt_opt.driver.options[mapped_keys[key]] = opt_options[key]
                else:
                    wt_opt.driver.options[key] = opt_options[key]

        # Loop through all of the opt_settings provided and set them in the OM driver object
        for key in opt_settings_keys:
            if key in opt_options:
                if key in mapped_keys:
                    wt_opt.driver.opt_settings[mapped_keys[key]] = opt_options[key]
                else:
                    wt_opt.driver.opt_settings[key] = opt_options[key]

        return wt_opt

    def set_driver(self, wt_opt):
        folder_output = self.opt["general"]["folder_output"]

        if self.opt["driver"]["optimization"]["flag"]:
            opt_options = self.opt["driver"]["optimization"]
            step_size = self._get_step_size()

            if opt_options["step_calc"] == "None":
                step_calc = None
            else:
                step_calc = opt_options["step_calc"]
            wt_opt.model.approx_totals(method="fd", step=step_size, form=opt_options["form"], step_calc=step_calc)

            # Set optimization solver and options. First, Scipy's SLSQP and COBYLA
            if opt_options["solver"] in self.scipy_methods:
                wt_opt.driver = om.ScipyOptimizeDriver()
                wt_opt.driver.options["optimizer"] = opt_options["solver"]

                options_keys = ["tol", "max_iter", "disp"]
                opt_settings_keys = ["rhobeg", "catol", "adaptive"]
                mapped_keys = {"max_iter": "maxiter"}
                wt_opt = self._set_optimizer_properties(wt_opt, options_keys, opt_settings_keys, mapped_keys)

            # The next two optimization methods require pyOptSparse.
            elif opt_options["solver"] in self.pyoptsparse_methods:
                try:
                    from openmdao.api import pyOptSparseDriver
                except:
                    raise ImportError(
                        f"You requested the optimization solver {opt_options['solver']}, but you have not installed pyOptSparse. Please do so and rerun."
                    )
                wt_opt.driver = pyOptSparseDriver()
                try:
                    wt_opt.driver.options["optimizer"] = opt_options["solver"]
                except:
                    raise ImportError(
                        f"You requested the optimization solver {opt_options['solver']}, but you have not installed it within pyOptSparse. Please build {opt_options['solver']} and rerun."
                    )

                # Most of the pyOptSparse options have special syntax when setting them,
                # so here we set them by hand instead of using `_set_optimizer_properties` for SNOPT and CONMIN.
                if opt_options["solver"] == "CONMIN":
                    wt_opt.driver.opt_settings["ITMAX"] = opt_options["max_iter"]

                if opt_options["solver"] == "NSGA2":
                    opt_settings_keys = [
                        "PopSize",
                        "maxGen",
                        "pCross_real",
                        "pMut_real",
                        "eta_c",
                        "eta_m",
                        "pCross_bin",
                        "pMut_bin",
                        "PrintOut",
                        "seed",
                        "xinit",
                    ]
                    wt_opt = self._set_optimizer_properties(wt_opt, opt_settings_keys=opt_settings_keys)

                elif opt_options["solver"] == "SNOPT":
                    wt_opt.driver.opt_settings["Major optimality tolerance"] = float(opt_options["tol"])
                    wt_opt.driver.opt_settings["Major iterations limit"] = int(opt_options["max_major_iter"])
                    wt_opt.driver.opt_settings["Iterations limit"] = int(opt_options["max_minor_iter"])
                    wt_opt.driver.opt_settings["Major feasibility tolerance"] = float(opt_options["tol"])
                    if "time_limit" in opt_options:
                        wt_opt.driver.opt_settings["Time limit"] = int(opt_options["time_limit"])
                    if "major_step_limit" in opt_options:
                        wt_opt.driver.opt_settings["Major step limit"] = float(opt_options["major_step_limit"])
                    if "function_precision" in opt_options:
                        wt_opt.driver.opt_settings["Function precision"] = float(opt_options["function_precision"])
                    if "verify_level" in opt_options:
                        wt_opt.driver.opt_settings["Verify level"] = opt_options["verify_level"]
                    else:
                        wt_opt.driver.opt_settings["Verify level"] = -1
                        
                # below are pyoptsparse options
                wt_opt.driver.options["output_dir"] = folder_output # Directory location of pyopt_sparse output files.Default is {prob_name}_out/reports. OpenMDAO overwrties SNOPT output file locations.
                if "hist_file_name" in opt_options:
                    wt_opt.driver.options["hist_file"] = os.path.join(folder_output, opt_options["hist_file_name"]) # File location for saving pyopt_sparse optimization history. Default is None for no output.
                if "hotstart_file" in opt_options:
                    wt_opt.driver.options["hotstart_file"] = opt_options["hotstart_file"] # File location of a pyopt_sparse optimization history to use to hot start the optimization. Default is None.

            elif opt_options["solver"] == "GA":
                wt_opt.driver = om.SimpleGADriver()
                options_keys = [
                    "Pc",
                    "Pm",
                    "bits",
                    "compute_pareto",
                    "cross_bits",
                    "elitism",
                    "gray",
                    "max_gen",
                    "multi_obj_exponent",
                    "multi_obj_weights",
                    "penalty_exponent",
                    "penalty_parameter",
                    "pop_size",
                    "procs_per_model",
                    "run_parallel",
                ]
                wt_opt = self._set_optimizer_properties(wt_opt, options_keys)

            elif opt_options["solver"] == "DE":
                wt_opt.driver = om.DifferentialEvolutionDriver()
                # wt_opt.driver.options["optimizer"] = opt_options["solver"]
                options_keys = [
                    "F", # differential rate
                    "Pc", # crossover probability
                    "debug_type", # what driver variables to print at each iteration
                    "invalid_desvar_behavior", # behavior of driver if initial DV violated bounds
                    "max_gen", # number of generations before termination
                    "multi_obj_exponent", # weighting exponent for multiobjective
                    "multi_obj_weights", # weights of objectives for multiobjective optimization (abs name -> weight dict)
                    "penalty_exponent", # penalty function exponent
                    "penalty_parameter", # penalty function parameter
                    "pop_size", # number of points in the GA (default to 20x inputs if 0)
                    "procs_per_model", # processors to give each model under mpi
                    "run_parallel", # true -> execut points in a generation in parallel
                ]
                mapped_keys = {"debug_print": "debug_type"}
                wt_opt = self._set_optimizer_properties(wt_opt, options_keys, mapped_keys=mapped_keys)

            elif opt_options["solver"] in self.nlopt_methods:
                try:
                    from wisdem.optimization_drivers.nlopt_driver import NLoptDriver
                except:
                    raise ImportError(
                        "You requested an optimization method from NLopt, but need to first install NLopt to use this method."
                    )

                wt_opt.driver = NLoptDriver()
                wt_opt.driver.options["optimizer"] = opt_options["solver"]
                options_keys = ["tol", "xtol", "max_iter", "max_time", "numgen"]
                mapped_keys = {"max_iter": "maxiter", "max_time": "maxtime"}
                wt_opt = self._set_optimizer_properties(wt_opt, options_keys, mapped_keys=mapped_keys)

            else:
                raise ValueError(f"The {self.opt['driver']['optimization']['solver']} optimizer is not yet supported!")

            if opt_options["debug_print"]:
                wt_opt.driver.options["debug_print"] = ["desvars", "ln_cons", "nl_cons", "objs", "totals"]

        elif self.opt["driver"]["design_of_experiments"]["flag"]:
            doe_options = self.opt["driver"]["design_of_experiments"]
            if doe_options["generator"].lower() == "uniform":
                generator = om.UniformGenerator(
                    num_samples=int(doe_options["num_samples"]),
                    seed=doe_options["seed"],
                )
            elif doe_options["generator"].lower() == "fullfact":
                generator = om.FullFactorialGenerator(levels=int(doe_options["num_samples"]))
            elif doe_options["generator"].lower() == "plackettburman":
                generator = om.PlackettBurmanGenerator()
            elif doe_options["generator"].lower() == "boxbehnken":
                generator = om.BoxBehnkenGenerator()
            elif doe_options["generator"].lower() == "latinhypercube":
                generator = om.LatinHypercubeGenerator(
                    samples=int(doe_options["num_samples"]),
                    criterion=doe_options["criterion"],
                    seed=doe_options["seed"],
                )
            else:
                raise Exception("The generator type {} is unsupported.".format(doe_options["generator"]))

            # Initialize driver
            wt_opt.driver = om.DOEDriver(generator)

            if doe_options["debug_print"]:
                wt_opt.driver.options["debug_print"] = ["desvars", "ln_cons", "nl_cons", "objs"]

            # options
            wt_opt.driver.options["run_parallel"] = doe_options["run_parallel"]

        elif self.opt["driver"]["step_size_study"]["flag"]:
            pass

        else:
            print("WARNING: Design variables are set to be optimized or studied, but no driver is selected.")
            print("         If you want to run an optimization, please enable a driver.")

        return wt_opt

    def set_objective(self, wt_opt):
        # Set merit figure. Each objective has its own scaling.  Check first for user override
        if self.opt["merit_figure_user"]["name"] != "":
            coeff = -1.0 if self.opt["merit_figure_user"]["max_flag"] else 1.0
            wt_opt.model.add_objective(self.opt["merit_figure_user"]["name"],
                                       ref=coeff*np.abs(self.opt["merit_figure_user"]["ref"]))

        elif self.opt["merit_figure"].lower() == "aep":
            wt_opt.model.add_objective("rotorse.rp.AEP", ref=-1.0e6)

        elif self.opt["merit_figure"] == "blade_mass":
            wt_opt.model.add_objective("rotorse.blade_mass", ref=1.0e6)

        elif self.opt["merit_figure"] == "blade_cost":
            wt_opt.model.add_objective("rotorse.total_bc.total_blade_cost", ref=1.0e6)

        elif self.opt["merit_figure"].lower() == "lcoe":
            wt_opt.model.add_objective("financese.lcoe", ref=0.1)

        elif self.opt["merit_figure"] == "blade_tip_deflection":
            wt_opt.model.add_objective("tcons.tip_deflection_ratio")

        elif self.opt["merit_figure"] == "tower_mass":
            wt_opt.model.add_objective("towerse.tower_mass", ref=1e6)

        elif self.opt["merit_figure"] == "tower_cost":
            wt_opt.model.add_objective("tcc.tower_cost", ref=1e6)
            
        elif self.opt["merit_figure"] == "monopile_mass":
            wt_opt.model.add_objective("fixedse.monopile_mass", ref=1e6)

        elif self.opt["merit_figure"] == "monopile_cost":
            wt_opt.model.add_objective("fixedse.monopile_cost", ref=1e6)

        elif self.opt["merit_figure"] == "jacket_mass":
            wt_opt.model.add_objective("fixedse.jacket_mass", ref=1e6)

        elif self.opt["merit_figure"] == "structural_mass":
            if not self.modeling["flags"]["floating"]:
                wt_opt.model.add_objective("fixedse.structural_mass", ref=1e6)
            else:
                wt_opt.model.add_objective("floatingse.system_structural_mass", ref=1e6)

        elif self.opt["merit_figure"] == "hub_mass":
            wt_opt.model.add_objective("drivese.hub_system_mass", ref=1e5)

        elif self.opt["merit_figure"] == "nacelle_mass":
            wt_opt.model.add_objective("drivese.nacelle_mass", ref=1e6)

        elif self.opt["merit_figure"] == "nacelle_cost":
            wt_opt.model.add_objective("tcc.nacelle_cost", ref=1e6)

        elif self.opt["merit_figure"] == "platform_hull_mass":
            wt_opt.model.add_objective("floatingse.platform_hull_mass", ref=1e6)

        elif self.opt["merit_figure"] == "platform_mass":
            wt_opt.model.add_objective("floatingse.platform_mass", ref=1e6)

        elif self.opt["merit_figure"] == "platform_cost":
            wt_opt.model.add_objective("floatingse.platform_cost", ref=1e6)

        elif self.opt["merit_figure"] == "mooring_mass":
            wt_opt.model.add_objective("floatingse.mooring_mass", ref=1e4)

        elif self.opt["merit_figure"] == "mooring_cost":
            wt_opt.model.add_objective("floatingse.mooring_cost", ref=1e4)

        elif self.opt["merit_figure"] == "Cp":
            if self.modeling["flags"]["blade"]:
                wt_opt.model.add_objective("rotorse.rp.powercurve.Cp_regII", ref=-1.0)
            else:
                wt_opt.model.add_objective("rotorse.ccblade.CP", ref=-1.0)

        elif self.opt["merit_figure"] in ["turbine_cost", "turbine_capex"]:
            wt_opt.model.add_objective("tcc.turbine_cost_kW", ref=1e3)

        elif self.opt["merit_figure"] in ["bos", "bos_cost", "bos_capex"]:
            if self.modeling["flags"]["offshore"]:
                wt_opt.model.add_objective("orbit.total_capex_kW", ref=1e3)
            else:
                wt_opt.model.add_objective("landbosse.total_capex_kW", ref=1e3)

        elif self.opt["merit_figure"] == "inverse_design":
            wt_opt.model.add_objective("inverse_design.objective")

        else:
            raise ValueError("The merit figure " + self.opt["merit_figure"] + " is unknown or not supported.")

        return wt_opt

    def set_design_variables(self, wt_opt, wt_init):
        # Set optimization design variables.
        rotorD_opt = self.opt["design_variables"]["rotor_diameter"]
        blade_opt = self.opt["design_variables"]["blade"]
        tower_opt = self.opt["design_variables"]["tower"]
        monopile_opt = self.opt["design_variables"]["monopile"]
        jacket_opt = self.opt["design_variables"]["jacket"]
        control_opt = self.opt["design_variables"]["control"]
        hub_opt = self.opt["design_variables"]["hub"]
        drive_opt = self.opt["design_variables"]["drivetrain"]
        float_opt = self.opt["design_variables"]["floating"]
        mooring_opt = self.opt["design_variables"]["mooring"]

        # -- Rotor & Blade --
        if rotorD_opt["flag"]:
            wt_opt.model.add_design_var(
                "configuration.rotor_diameter_user", lower=rotorD_opt["minimum"], upper=rotorD_opt["maximum"], ref=1.0e2
            )

        twist_options = blade_opt["aero_shape"]["twist"]
        if twist_options["flag"]:
            if blade_opt["aero_shape"]["twist"]["index_end"] > blade_opt["aero_shape"]["twist"]["n_opt"]:
                raise Exception(
                    "Check the analysis options yaml, index_end of the blade twist is higher than the number of DVs n_opt"
                )
            elif blade_opt["aero_shape"]["twist"]["index_end"] == 0:
                blade_opt["aero_shape"]["twist"]["index_end"] = blade_opt["aero_shape"]["twist"]["n_opt"]
            indices_twist = range(twist_options["index_start"], twist_options["index_end"])
            s_opt_twist = np.linspace(0.0, 1.0, blade_opt["aero_shape"]["twist"]["n_opt"])
            twist_interpolator = PchipInterpolator(
                wt_init["components"]["blade"]["outer_shape_bem"]["twist"]["grid"],
                wt_init["components"]["blade"]["outer_shape_bem"]["twist"]["values"],
            )
            init_twist_opt = twist_interpolator(s_opt_twist)
            wt_opt.model.add_design_var(
                "blade.opt_var.twist_opt",
                indices=indices_twist,
                lower=init_twist_opt[indices_twist] - blade_opt["aero_shape"]["twist"]["max_decrease"],
                upper=init_twist_opt[indices_twist] + blade_opt["aero_shape"]["twist"]["max_increase"],
            )

        chord_options = blade_opt["aero_shape"]["chord"]
        if chord_options["flag"]:
            if blade_opt["aero_shape"]["chord"]["index_end"] > blade_opt["aero_shape"]["chord"]["n_opt"]:
                raise Exception(
                    "Check the analysis options yaml, index_end of the blade chord is higher than the number of DVs n_opt"
                )
            elif blade_opt["aero_shape"]["chord"]["index_end"] == 0:
                blade_opt["aero_shape"]["chord"]["index_end"] = blade_opt["aero_shape"]["chord"]["n_opt"]
            indices_chord = range(chord_options["index_start"], chord_options["index_end"])
            s_opt_chord = np.linspace(0.0, 1.0, blade_opt["aero_shape"]["chord"]["n_opt"])
            chord_interpolator = PchipInterpolator(
                wt_init["components"]["blade"]["outer_shape_bem"]["chord"]["grid"],
                wt_init["components"]["blade"]["outer_shape_bem"]["chord"]["values"],
            )
            init_chord_opt = chord_interpolator(s_opt_chord)
            wt_opt.model.add_design_var(
                "blade.opt_var.chord_opt",
                indices=indices_chord,
                lower=init_chord_opt[indices_chord] * chord_options["max_decrease"],
                upper=init_chord_opt[indices_chord] * chord_options["max_increase"],
            )

        if blade_opt["aero_shape"]["af_positions"]["flag"]:
            n_af = self.modeling["WISDEM"]["RotorSE"]["n_af_span"]
            indices_af = range(blade_opt["aero_shape"]["af_positions"]["af_start"], n_af - 1)
            af_pos_init = wt_init["components"]["blade"]["outer_shape_bem"]["airfoil_position"]["grid"]
            step_size = self._get_step_size()
            lb_af = np.zeros(n_af)
            ub_af = np.zeros(n_af)
            for i in range(1, indices_af[0]):
                lb_af[i] = ub_af[i] = af_pos_init[i]
            for i in indices_af:
                lb_af[i] = 0.5 * (af_pos_init[i - 1] + af_pos_init[i]) + step_size
                ub_af[i] = 0.5 * (af_pos_init[i + 1] + af_pos_init[i]) - step_size
            lb_af[-1] = ub_af[-1] = 1.0
            wt_opt.model.add_design_var(
                "blade.opt_var.af_position", indices=indices_af, lower=lb_af[indices_af], upper=ub_af[indices_af]
            )

        L_D_options = blade_opt["aero_shape"]["L/D"]
        if L_D_options["flag"]:
            n_opt = L_D_options["n_opt"]
            indices = range(L_D_options["index_start"], L_D_options["index_end"])
            s_opt_L_D = np.linspace(0.0, 1.0, n_opt)
            L_D_interpolator = PchipInterpolator(
                wt_init["components"]["blade"]["outer_shape_bem"]["L/D"]["grid"],
                wt_init["components"]["blade"]["outer_shape_bem"]["L/D"]["values"],
            )
            init_L_D_opt = L_D_interpolator(s_opt_L_D)

            wt_opt.model.add_design_var(
                "inn_af.L_D_opt",
                indices=indices,
                lower=init_L_D_opt[indices] - L_D_options["max_decrease"],
                upper=init_L_D_opt[indices] + L_D_options["max_increase"],
            )

        c_d_options = blade_opt["aero_shape"]["c_d"]
        if c_d_options["flag"]:
            n_opt = c_d_options["n_opt"]
            indices = range(c_d_options["index_start"], c_d_options["index_end"])
            s_opt_c_d = np.linspace(0.0, 1.0, n_opt)
            c_d_interpolator = PchipInterpolator(
                wt_init["components"]["blade"]["outer_shape_bem"]["c_d"]["grid"],
                wt_init["components"]["blade"]["outer_shape_bem"]["c_d"]["values"],
            )
            init_c_d_opt = c_d_interpolator(s_opt_c_d)

            wt_opt.model.add_design_var(
                "inn_af.c_d_opt",
                indices=indices,
                lower=init_c_d_opt[indices] - c_d_options["max_decrease"],
                upper=init_c_d_opt[indices] + c_d_options["max_increase"],
            )

        stall_options = blade_opt["aero_shape"]["stall_margin"]
        if stall_options["flag"]:
            n_opt = stall_options["n_opt"]
            indices = range(stall_options["index_start"], stall_options["index_end"])
            s_opt_stall = np.linspace(0.0, 1.0, n_opt)
            stall_interpolator = PchipInterpolator(
                wt_init["components"]["blade"]["outer_shape_bem"]["stall_margin"]["grid"],
                wt_init["components"]["blade"]["outer_shape_bem"]["stall_margin"]["values"],
            )
            init_stall_opt = stall_interpolator(s_opt_stall)

            wt_opt.model.add_design_var(
                "inn_af.stall_margin_opt",
                indices=indices,
                lower=init_stall_opt[indices] - stall_options["max_decrease"],
                upper=init_stall_opt[indices] + stall_options["max_increase"],
            )

        t_c_options = blade_opt["aero_shape"]["rthick"]
        if t_c_options["flag"]:
            n_opt = t_c_options["n_opt"]
            indices = range(t_c_options["index_start"], t_c_options["index_end"])
            s_opt_t_c = np.linspace(0.0, 1.0, n_opt)
            t_c_interpolator = PchipInterpolator(
                wt_init["components"]["blade"]["outer_shape_bem"]["rthick"]["grid"],
                wt_init["components"]["blade"]["outer_shape_bem"]["rthick"]["values"],
            )
            init_t_c_opt = t_c_interpolator(s_opt_t_c)

            wt_opt.model.add_design_var(
                "inn_af.r_thick_opt",
                indices=indices,
                lower=init_t_c_opt[indices] - t_c_options["max_decrease"],
                upper=init_t_c_opt[indices] + t_c_options["max_increase"],
            )

        z_options = blade_opt["aero_shape"]["z"]
        if z_options["flag"]:
            wt_opt.model.add_design_var(
                "inn_af.z",
                lower=z_options["lower_bound"],
                upper=z_options["upper_bound"],
            )

        if "structure" in blade_opt and len(blade_opt["structure"]) > 0:
            if self.modeling["user_elastic"]["blade"]:
                raise Exception("Blade structural design variables not available for user-defined blade elastic model. Please modify the modeling or optimization options.")
            layers = wt_init["components"]["blade"]["internal_structure_2d_fem"]["layers"]
            for i in range(len(blade_opt["structure"])):
                k = blade_opt["layer_index_opt"][i]
                layer_interp = PchipInterpolator(
                        layers[k]["thickness"]["grid"],
                        layers[k]["thickness"]["values"],
                        extrapolate=False
                    )
                s_opt_layer = np.linspace(0.0, 1.0, blade_opt["n_opt_struct"][k])
                init_layer_opt = layer_interp(s_opt_layer)
                # Only set an optimization variable if the layer has non zero initial thickness
                indices_i = [
                    idx for idx in range(blade_opt["structure"][i]["index_start"], blade_opt["structure"][i]["index_end"])
                    if init_layer_opt[idx] >= 1.e-6
                ]
                wt_opt.model.add_design_var(
                    "blade.opt_var.layer_%d_opt"%k,
                    indices=indices_i,
                    lower=init_layer_opt[indices_i] * blade_opt["structure"][i]["max_decrease"],
                    upper=init_layer_opt[indices_i] * blade_opt["structure"][i]["max_increase"],
                    ref=1.0e-2,
                )

        # -- Tower --
        if tower_opt["outer_diameter"]["flag"]:
            wt_opt.model.add_design_var(
                "tower.diameter",
                lower=tower_opt["outer_diameter"]["lower_bound"],
                upper=tower_opt["outer_diameter"]["upper_bound"],
                ref=5.0,
            )

        if tower_opt["layer_thickness"]["flag"]:
            wt_opt.model.add_design_var(
                "tower.layer_thickness",
                lower=tower_opt["layer_thickness"]["lower_bound"],
                upper=tower_opt["layer_thickness"]["upper_bound"],
                ref=1e-2,
            )

        if tower_opt["E"]["flag"]:
            ivc = wt_opt.model.add_subsystem("E_ivc", om.IndepVarComp(), promotes=[])
            ivc.add_output("E_user", val=10.0, units="Pa")
            wt_opt.model.add_design_var(
                "E_ivc.E_user",
                lower=tower_opt["E"]["lower_bound"],
                upper=tower_opt["E"]["upper_bound"],
                ref=1e9,
            )
            wt_opt.model.connect("E_ivc.E_user", "towerse.E_user")

            if self.modeling["flags"]["monopile"] or self.modeling["flags"]["jacket"]:
                wt_opt.model.connect("E_ivc.E_user", "fixedse.E_user")

        for idx, material in enumerate(wt_init["materials"]):
            if material["name"] == "steel":
                tower_material_index = idx

        if tower_opt["rho"]["flag"]:
            wt_opt.model.add_design_var(
                "materials.rho",
                lower=tower_opt["rho"]["lower_bound"],
                upper=tower_opt["rho"]["upper_bound"],
                ref=1e3,
                indices=[tower_material_index],
            )

        # -- Monopile --
        if monopile_opt["outer_diameter"]["flag"]:
            wt_opt.model.add_design_var(
                "monopile.diameter",
                lower=monopile_opt["outer_diameter"]["lower_bound"],
                upper=monopile_opt["outer_diameter"]["upper_bound"],
                ref=5.0,
            )

        if monopile_opt["layer_thickness"]["flag"]:
            wt_opt.model.add_design_var(
                "monopile.layer_thickness",
                lower=monopile_opt["layer_thickness"]["lower_bound"],
                upper=monopile_opt["layer_thickness"]["upper_bound"],
                ref=1e-2,
            )

        # -- Jacket --
        if jacket_opt["foot_head_ratio"]["flag"]:
            wt_opt.model.add_design_var(
                "jacket.foot_head_ratio",
                lower=jacket_opt["foot_head_ratio"]["lower_bound"],
                upper=jacket_opt["foot_head_ratio"]["upper_bound"],
            )

        if jacket_opt["r_head"]["flag"]:
            wt_opt.model.add_design_var(
                "jacket.r_head",
                lower=jacket_opt["r_head"]["lower_bound"],
                upper=jacket_opt["r_head"]["upper_bound"],
                ref=5.0,
            )

        if jacket_opt["leg_diameter"]["flag"]:
            wt_opt.model.add_design_var(
                "jacket.leg_diameter",
                lower=jacket_opt["leg_diameter"]["lower_bound"],
                upper=jacket_opt["leg_diameter"]["upper_bound"],
            )

        if jacket_opt["leg_thickness"]["flag"]:
            wt_opt.model.add_design_var(
                "jacket.leg_thickness",
                lower=jacket_opt["leg_thickness"]["lower_bound"],
                upper=jacket_opt["leg_thickness"]["upper_bound"],
            )

        if jacket_opt["brace_diameters"]["flag"]:
            wt_opt.model.add_design_var(
                "jacket.brace_diameters",
                lower=jacket_opt["brace_diameters"]["lower_bound"],
                upper=jacket_opt["brace_diameters"]["upper_bound"],
            )

        if jacket_opt["brace_thicknesses"]["flag"]:
            wt_opt.model.add_design_var(
                "jacket.brace_thicknesses",
                lower=jacket_opt["brace_thicknesses"]["lower_bound"],
                upper=jacket_opt["brace_thicknesses"]["upper_bound"],
            )

        if jacket_opt["height"]["flag"]:
            wt_opt.model.add_design_var(
                "jacket.height",
                lower=jacket_opt["height"]["lower_bound"],
                upper=jacket_opt["height"]["upper_bound"],
            )

        if jacket_opt["bay_spacing"]["flag"]:
            wt_opt.model.add_design_var(
                "jacket.bay_spacing",
                lower=jacket_opt["bay_spacing"]["lower_bound"],
                upper=jacket_opt["bay_spacing"]["upper_bound"],
            )

        # -- Control --
        if control_opt["tsr"]["flag"]:
            wt_opt.model.add_design_var(
                "control.rated_TSR", lower=control_opt["tsr"]["minimum"], upper=control_opt["tsr"]["maximum"], ref=1e1
            )
        # -- Hub & Drivetrain --
        if hub_opt["cone"]["flag"]:
            wt_opt.model.add_design_var(
                "hub.cone", lower=hub_opt["cone"]["lower_bound"], upper=hub_opt["cone"]["upper_bound"], ref=1e-2
            )
        if hub_opt["hub_diameter"]["flag"]:
            wt_opt.model.add_design_var(
                "hub.diameter",
                lower=hub_opt["hub_diameter"]["lower_bound"],
                upper=hub_opt["hub_diameter"]["upper_bound"],
            )
        if drive_opt["uptilt"]["flag"]:
            wt_opt.model.add_design_var(
                "nacelle.uptilt",
                lower=drive_opt["uptilt"]["lower_bound"],
                upper=drive_opt["uptilt"]["upper_bound"],
                ref=1e-2,
            )

        if drive_opt["generator_length"]["flag"]:
            wt_opt.model.add_design_var(
                "nacelle.L_generator",
                lower=drive_opt["generator_length"]["lower_bound"],
                upper=drive_opt["generator_length"]["upper_bound"],
            )

        for k in [
            "overhang",
            "distance_tt_hub",
            "distance_hub_mb",
            "distance_mb_mb",
            "gear_ratio",
            "bedplate_flange_width",
            "lss_diameter",
            "hss_diameter",
            "nose_diameter",
        ]:
            if drive_opt[k]["flag"]:
                wt_opt.model.add_design_var(
                    "nacelle." + k, lower=drive_opt[k]["lower_bound"], upper=drive_opt[k]["upper_bound"]
                )

        for k in [
            "bedplate_wall_thickness",
            "bedplate_web_thickness",
            "bedplate_flange_thickness",
            "lss_wall_thickness",
            "hss_wall_thickness",
            "nose_wall_thickness",
        ]:
            if drive_opt[k]["flag"]:
                wt_opt.model.add_design_var(
                    "nacelle." + k, lower=drive_opt[k]["lower_bound"], upper=drive_opt[k]["upper_bound"], ref=1e-2
                )

        # -- Floating --
        if float_opt["joints"]["flag"]:
            jointz = float_opt["joints"]["z_coordinate"]
            jointr = float_opt["joints"]["r_coordinate"]

            count = 0
            for k in range(len(jointz)):
                wt_opt.model.add_design_var(
                    f"floating.jointdv_{count}",
                    lower=jointz[k]["lower_bound"],
                    upper=jointz[k]["upper_bound"],
                )
                count += 1

            for k in range(len(jointr)):
                wt_opt.model.add_design_var(
                    f"floating.jointdv_{count}",
                    lower=jointr[k]["lower_bound"],
                    upper=jointr[k]["upper_bound"],
                )
                count += 1

        if float_opt["members"]["flag"]:
            for kgrp in float_opt["members"]["groups"]:
                memname = kgrp["names"][0]
                idx = self.modeling["floating"]["members"]["name2idx"][memname]
                imem = self.modeling["floating"]["members"]["name"].index(memname)
                istruct = wt_init["components"]["floating_platform"]["members"][imem]["internal_structure"]

                if "diameter" in kgrp:
                    wt_opt.model.add_design_var(
                        f"floating.memgrp{idx}.outer_diameter_in",
                        lower=kgrp["diameter"]["lower_bound"],
                        upper=kgrp["diameter"]["upper_bound"],
                    )
                if "side_length_a" in kgrp:
                    wt_opt.model.add_design_var(
                        f"floating.memgrp{idx}.side_length_a_in",
                        lower=kgrp["side_length_a"]["lower_bound"],
                        upper=kgrp["side_length_a"]["upper_bound"],
                    )
                    wt_opt.model.add_design_var(
                        f"floating.memgrp{idx}.side_length_b_in",
                        lower=kgrp["side_length_b"]["lower_bound"],
                        upper=kgrp["side_length_b"]["upper_bound"],
                    )
                if "thickness" in kgrp:
                    wt_opt.model.add_design_var(
                        f"floating.memgrp{idx}.layer_thickness_in",
                        lower=kgrp["thickness"]["lower_bound"],
                        upper=kgrp["thickness"]["upper_bound"],
                    )
                if "ballast" in kgrp and len(istruct["ballasts"]) > 0:
                    V_ballast = np.zeros(len(istruct["ballasts"]))
                    for j in range(V_ballast.size):
                        if "volume" in istruct["ballasts"][j]:
                            V_ballast[j] = istruct["ballasts"][j]["volume"]
                    iball = np.where(V_ballast > 0.0)[0]
                    if iball.size > 0:
                        wt_opt.model.add_design_var(
                            f"floating.memgrp{idx}.ballast_volume",
                            lower=kgrp["ballast"]["lower_bound"],
                            upper=kgrp["ballast"]["upper_bound"],
                            indices=iball,
                            ref=1e3,
                        )
                if "stiffeners" in kgrp:
                    if "ring" in kgrp["stiffeners"]:
                        if "size" in kgrp["stiffeners"]["ring"]:
                            pass
                        if "spacing" in kgrp["stiffeners"]["ring"]:
                            wt_opt.model.add_design_var(
                                f"floating.memgrp{idx}.ring_stiffener_spacing",
                                lower=kgrp["stiffeners"]["ring"]["spacing"]["lower_bound"],
                                upper=kgrp["stiffeners"]["ring"]["spacing"]["upper_bound"],
                            )
                    if "longitudinal" in kgrp["stiffeners"]:
                        if "size" in kgrp["stiffeners"]["longitudinal"]:
                            pass
                        if "spacing" in kgrp["stiffeners"]["longitudinal"]:
                            wt_opt.model.add_design_var(
                                f"floating.memgrp{idx}.axial_stiffener_spacing",
                                lower=kgrp["stiffeners"]["longitudinal"]["spacing"]["lower_bound"],
                                upper=kgrp["stiffeners"]["longitudinal"]["spacing"]["upper_bound"],
                            )
                if "axial_joints" in kgrp:
                    aidx = []
                    lower = []
                    upper = []
                    mem_axial_names = self.modeling["floating"]["members"]["axial_joint_name_member_" + memname]
                    for agrp in kgrp["axial_joints"]:
                        tryidx = None
                        agrp_names = agrp["names"]
                        lower.append(agrp["lower_bound"])
                        upper.append(agrp["upper_bound"])
                        for iname in agrp_names:
                            try:
                                tryidx = mem_axial_names.index(iname)
                                aidx.append(tryidx)
                                break
                            except Exception:
                                continue
                        if tryidx is None:
                            raise ValueError(
                                f"None of these axial joint names were found in member, {memname}:{agrp_names}"
                            )
                    wt_opt.model.add_design_var(
                        f"floating.memgrp{idx}.grid_axial_joints", lower=lower, upper=upper, indices=aidx
                    )

        # -- Mooring --
        if mooring_opt["line_length"]["flag"]:
            wt_opt.model.add_design_var(
                "mooring.unstretched_length_in",
                lower=mooring_opt["line_length"]["lower_bound"],
                upper=mooring_opt["line_length"]["upper_bound"],
                ref=1e2,
            )

        if mooring_opt["line_diameter"]["flag"]:
            wt_opt.model.add_design_var(
                "mooring.line_diameter_in",
                lower=mooring_opt["line_diameter"]["lower_bound"],
                upper=mooring_opt["line_diameter"]["upper_bound"],
                ref=1e-1,
            )

        if mooring_opt["line_mass_density_coeff"]["flag"]:
            wt_opt.model.add_design_var(
                "mooring.line_mass_density_coeff",
                lower=mooring_opt["line_mass_density_coeff"]["lower_bound"],
                upper=mooring_opt["line_mass_density_coeff"]["upper_bound"],
            )

        if mooring_opt["line_stiffness_coeff"]["flag"]:
            wt_opt.model.add_design_var(
                "mooring.line_stiffness_coeff",
                lower=mooring_opt["line_stiffness_coeff"]["lower_bound"],
                upper=mooring_opt["line_stiffness_coeff"]["upper_bound"],
            )

        if "user" in self.opt["design_variables"]:
            user_defined = self.opt["design_variables"]["user"]
            for i in range(len(user_defined)):
                name_i = user_defined[i]["name"]

                if "lower_bound" in user_defined[i]:
                    lower_i = user_defined[i]["lower_bound"]
                elif "lower" in user_defined[i]:
                    lower_i = user_defined[i]["lower"]
                else:
                    lower_i = None

                if "upper_bound" in user_defined[i]:
                    upper_i = user_defined[i]["upper_bound"]
                elif "upper" in user_defined[i]:
                    upper_i = user_defined[i]["upper"]
                else:
                    upper_i = None

                if "ref" in user_defined[i]:
                    ref_i = user_defined[i]["ref"]
                else:
                    ref_i = None

                if "indices" in user_defined[i]:
                    indices_i = user_defined[i]["indices"]
                else:
                    indices_i = None

                wt_opt.model.add_design_var(
                    name_i,
                    lower=lower_i,
                    upper=upper_i,
                    ref=ref_i,
                    indices=indices_i,
                )

        return wt_opt

    def set_constraints(self, wt_opt):
        blade_opt = self.opt["design_variables"]["blade"]

        # Set non-linear blade constraints
        blade_constr = self.opt["constraints"]["blade"]

        if blade_constr["strains_spar_cap_ss"]["flag"]:
            indices_strains_spar_cap_ss = range(
                    blade_constr["strains_spar_cap_ss"]["index_start"], blade_constr["strains_spar_cap_ss"]["index_end"]
                )
            wt_opt.model.add_constraint(
                    "rotorse.rs.constr.constr_max_strainU_spar", indices=indices_strains_spar_cap_ss, upper=1.0
                )
        if blade_constr["strains_spar_cap_ps"]["flag"]:
            indices_strains_spar_cap_ps = range(
                    blade_constr["strains_spar_cap_ps"]["index_start"], blade_constr["strains_spar_cap_ps"]["index_end"]
                )
            wt_opt.model.add_constraint(
                    "rotorse.rs.constr.constr_max_strainL_spar", indices=indices_strains_spar_cap_ps, upper=1.0
                )
        if blade_constr["strains_te_ss"]["flag"]:
            indices_strains_te_ss = range(
                    blade_constr["strains_te_ss"]["index_start"], blade_constr["strains_te_ss"]["index_end"]
                )
            wt_opt.model.add_constraint(
                    "rotorse.rs.constr.constr_max_strainU_te", indices=indices_strains_te_ss, upper=1.0
                )
        if blade_constr["strains_te_ps"]["flag"]:
            indices_strains_te_ps = range(
                    blade_constr["strains_te_ps"]["index_start"], blade_constr["strains_te_ps"]["index_end"]
                )
            wt_opt.model.add_constraint(
                    "rotorse.rs.constr.constr_max_strainL_te", indices=indices_strains_te_ps, upper=1.0
                )
        if blade_constr["stall"]["flag"]:
            wt_opt.model.add_constraint("rotorse.stall_check.no_stall_constraint", upper=1.0)
            if not blade_opt["aero_shape"]["twist"]["flag"]:
                print(
                    "WARNING: the margin to stall is set to be constrained, but twist is not an active design variable."
                )
        if blade_constr["tip_deflection"]["flag"]:
            wt_opt.model.add_constraint("tcons.tip_deflection_ratio", upper=1.0)
        if blade_constr["t_sc_joint"]["flag"] and self.modeling["WISDEM"]["RotorSE"]["bjs"]:
            wt_opt.model.add_constraint(
                "rotorse.rs.bjs.t_sc_ratio_joint", upper=1.0
            )
        if blade_constr["chord"]["flag"]:
            if blade_opt["aero_shape"]["chord"]["flag"]:
                wt_opt.model.add_constraint("blade.pa.max_chord_constr", upper=1.0)
            else:
                print(
                    "WARNING: the max chord is set to be constrained, but chord is not an active design variable. The constraint is not enforced."
                )
        if blade_constr["chord_slope"]["flag"]:
            if blade_opt["aero_shape"]["chord"]["flag"]:
                wt_opt.model.add_constraint("blade.pa.slope_chord_constr", upper=0.0)
            else:
                print(
                    "WARNING: the slope of the chord is set to be constrained, but chord is not an active design variable. The constraint is not enforced."
                )
        if blade_constr["twist_slope"]["flag"]:
            if blade_opt["aero_shape"]["twist"]["flag"]:
                wt_opt.model.add_constraint("blade.pa.slope_twist_constr", upper=0.0)
            else:
                print(
                    "WARNING: the slope of the twist is set to be constrained, but twist is not an active design variable. The constraint is not enforced."
                )

        if blade_constr["root_circle_diameter"]["flag"]:
            if blade_opt["aero_shape"]["chord"]["flag"] and blade_opt["aero_shape"]["chord"]["index_start"] == 0.0:
                wt_opt.model.add_constraint(
                    "rotorse.rs.brs.ratio", upper=blade_constr["root_circle_diameter"]["max_ratio"]
                )
            else:
                print(
                    "WARNING: the blade root size is set to be constrained, but chord at blade root is not an active design variable. The constraint is not enforced."
                )
        if blade_constr["frequency"]["flap_3P"]:
            wt_opt.model.add_constraint("rotorse.rs.constr.constr_flap_f_margin", upper=0.0)
        if blade_constr["frequency"]["edge_3P"]:
            wt_opt.model.add_constraint("rotorse.rs.constr.constr_edge_f_margin", upper=0.0)
        if blade_constr["frequency"]["first_flap"]["flag"]:
            target = blade_constr["frequency"]["first_flap"]["target"]
            error = blade_constr["frequency"]["first_flap"]["acceptable_error"]
            upper_value = target + error
            lower_value = target - error
            wt_opt.model.add_constraint("rotorse.rs.frame.flap_mode_freqs", indices=[0], upper = upper_value, lower = lower_value)
        # if blade_constr["frequency"]["second_flap"]["flag"]:
        #     target = blade_constr["frequency"]["second_flap"]["target"]
        #     error = blade_constr["frequency"]["second_flap"]["acceptable_error"]
        #     upper_value = target + error
        #     lower_value = target - error
        #     wt_opt.model.add_constraint("rotorse.rs.frame.flap_mode_freqs", indices=[1], upper = upper_value, lower = lower_value, alias = '_2F')
        if blade_constr["frequency"]["first_edge"]["flag"]:
            target = blade_constr["frequency"]["first_edge"]["target"]
            error = blade_constr["frequency"]["first_edge"]["acceptable_error"]
            upper_value = target + error
            lower_value = target - error
            wt_opt.model.add_constraint("rotorse.rs.frame.edge_mode_freqs", indices=[0], upper = upper_value, lower = lower_value)
        # if blade_constr["frequency"]["second_edge"]["flag"]:
        #     target = blade_constr["frequency"]["second_edge"]["target"]
        #     error = blade_constr["frequency"]["second_edge"]["acceptable_error"]
        #     upper_value = target + error
        #     lower_value = target - error
        #     wt_opt.model.add_constraint("rotorse.rs.frame.edge_mode_freqs", indices=[1], upper = upper_value, lower = lower_value, alias = '_2E')
        if blade_constr["frequency"]["first_torsion"]["flag"]:
            target = blade_constr["frequency"]["first_torsion"]["target"]
            error = blade_constr["frequency"]["first_torsion"]["acceptable_error"]
            upper_value = target + error
            lower_value = target - error
            wt_opt.model.add_constraint("rotorse.rs.frame.tors_mode_freqs", indices=[0], upper = upper_value, lower = lower_value)
        if blade_constr["mass"]["flag"]:
            target = blade_constr["mass"]["target"]
            error = blade_constr["mass"]["acceptable_error"]
            upper_value = target + error
            lower_value = target - error
            wt_opt.model.add_constraint("rotorse.blade_mass", upper = upper_value, lower = lower_value, ref = 1.e+5)
        if blade_constr["rail_transport"]["8_axle"]:
            wt_opt.model.add_constraint("rotorse.re.rail.constr_LV_8axle_horiz", lower=0.8, upper=1.0)
            wt_opt.model.add_constraint("rotorse.re.rail.constr_strainPS", upper=1.0)
            wt_opt.model.add_constraint("rotorse.re.rail.constr_strainSS", upper=1.0)
        elif blade_constr["rail_transport"]["4_axle"]:
            wt_opt.model.add_constraint("rotorse.re.rail.constr_LV_4axle_horiz", upper=1.0)
            wt_opt.model.add_constraint("rotorse.re.rail.constr_strainPS", upper=1.0)
            wt_opt.model.add_constraint("rotorse.re.rail.constr_strainSS", upper=1.0)

        if blade_constr["rated_velocity"]["flag"]:
            target = blade_constr["rated_velocity"]["target"]
            error = blade_constr["rated_velocity"]["acceptable_error"]
            upper_value = target + error
            lower_value = target - error
            wt_opt.model.add_constraint("rotorse.rp.powercurve.rated_V", upper = upper_value, lower = lower_value, ref = 10.0)
            
        if self.opt["constraints"]["blade"]["moment_coefficient"]["flag"]:
            wt_opt.model.add_constraint(
                "rotorse.ccblade.CM",
                lower=self.opt["constraints"]["blade"]["moment_coefficient"]["min"],
                upper=self.opt["constraints"]["blade"]["moment_coefficient"]["max"],
            )
        if (
            self.opt["constraints"]["blade"]["match_cl_cd"]["flag_cl"]
            or self.opt["constraints"]["blade"]["match_cl_cd"]["flag_cd"]
        ):
            data_target = np.loadtxt(self.opt["constraints"]["blade"]["match_cl_cd"]["filename"])
            eta_opt = np.linspace(0.0, 1.0, self.opt["design_variables"]["blade"]["aero_shape"]["twist"]["n_opt"])
            target_cl = PchipInterpolator(data_target[:, 0], data_target[:, 3])(eta_opt)
            target_cd = PchipInterpolator(data_target[:, 0], data_target[:, 4])(eta_opt)
            eps_cl = 1.0e-2
            if self.opt["constraints"]["blade"]["match_cl_cd"]["flag_cl"]:
                wt_opt.model.add_constraint(
                    "rotorse.ccblade.cl_n_opt", lower=target_cl - eps_cl, upper=target_cl + eps_cl
                )
            if self.opt["constraints"]["blade"]["match_cl_cd"]["flag_cd"]:
                wt_opt.model.add_constraint(
                    "rotorse.ccblade.cd_n_opt", lower=target_cd - eps_cl, upper=target_cd + eps_cl
                )
        if (
            self.opt["constraints"]["blade"]["match_L_D"]["flag_L"]
            or self.opt["constraints"]["blade"]["match_L_D"]["flag_D"]
        ):
            data_target = np.loadtxt(self.opt["constraints"]["blade"]["match_L_D"]["filename"])
            eta_opt = np.linspace(0.0, 1.0, self.opt["design_variables"]["blade"]["aero_shape"]["twist"]["n_opt"])
            target_L = PchipInterpolator(data_target[:, 0], data_target[:, 7])(eta_opt)
            target_D = PchipInterpolator(data_target[:, 0], data_target[:, 8])(eta_opt)
        eps_L = 1.0e2
        if self.opt["constraints"]["blade"]["match_L_D"]["flag_L"]:
            wt_opt.model.add_constraint("rotorse.ccblade.L_n_opt", lower=target_L - eps_L, upper=target_L + eps_L)
        if self.opt["constraints"]["blade"]["match_L_D"]["flag_D"]:
            wt_opt.model.add_constraint("rotorse.ccblade.D_n_opt", lower=target_D - eps_L, upper=target_D + eps_L)

        if self.opt["constraints"]["blade"]["AEP"]["flag"]:
            wt_opt.model.add_constraint(
                "rotorse.rp.AEP", lower=self.opt["constraints"]["blade"]["AEP"]["min"], ref=-1.0e6
            )

        # to constrain C_T and forcibly design higher thrust turbines
        if blade_constr["thrust_coeff"]["flag"]:
            if "lower_bound" in blade_constr["thrust_coeff"]:
                wt_opt.model.add_constraint(
                    "rotorse.rp.powercurve.Ct_regII", lower=blade_constr["thrust_coeff"]["lower_bound"]
                )
            elif "upper_bound" in blade_constr["thrust_coeff"]:
                wt_opt.model.add_constraint(
                    "rotorse.rp.powercurve.Ct_regII", upper=blade_constr["thrust_coeff"]["upper_bound"]
                )
            else:
                raise Exception("thrust coefficient constraint requested but no upper or lower bound found.")

        # Tower and monopile contraints
        tower_constr = self.opt["constraints"]["tower"]
        monopile_constr = self.opt["constraints"]["monopile"]

        if tower_constr["height_constraint"]["flag"]:
            wt_opt.model.add_constraint(
                "towerse.height_constraint",
                lower=tower_constr["height_constraint"]["lower_bound"],
                upper=tower_constr["height_constraint"]["upper_bound"],
            )

        if tower_constr["stress"]["flag"] and not monopile_constr["stress"]["flag"]:
            wt_opt.model.add_constraint("towerse.post.constr_stress", upper=1.0)

        if tower_constr["global_buckling"]["flag"] and not monopile_constr["global_buckling"]["flag"]:
            wt_opt.model.add_constraint("towerse.post.constr_global_buckling", upper=1.0)

        if tower_constr["shell_buckling"]["flag"] and not monopile_constr["shell_buckling"]["flag"]:
            wt_opt.model.add_constraint("towerse.post.constr_shell_buckling", upper=1.0)

        if tower_constr["d_to_t"]["flag"]:
            wt_opt.model.add_constraint(
                "towerse.constr_d_to_t",
                lower=tower_constr["d_to_t"]["lower_bound"],
                upper=tower_constr["d_to_t"]["upper_bound"],
            )

        if tower_constr["taper"]["flag"]:
            wt_opt.model.add_constraint("towerse.constr_taper", lower=tower_constr["taper"]["lower_bound"])

        if tower_constr["slope"]["flag"]:
            wt_opt.model.add_constraint("towerse.slope", upper=1.0)

        if tower_constr["thickness_slope"]["flag"]:
            wt_opt.model.add_constraint("towerse.thickness_slope", upper=1.0)

        if tower_constr["frequency"]["flag"]:
            wt_opt.model.add_constraint("tcons.constr_tower_f_1Pmargin", upper=0.0)
            wt_opt.model.add_constraint("tcons.constr_tower_f_NPmargin", upper=0.0)

        elif tower_constr["frequency_1"]["flag"]:
            # Cannot set both 1P/3P and hard freq values, so else-if statement here
            varstr = (
                "floatingse.structural_frequencies"
                if self.modeling["flags"]["floating"]
                else "towerse.tower.structural_frequencies"
            )
            wt_opt.model.add_constraint(
                varstr,
                indices=[0],
                lower=tower_constr["frequency_1"]["lower_bound"],
                upper=tower_constr["frequency_1"]["upper_bound"],
            )

        if tower_constr["mass"]["flag"]:
            target = tower_constr["mass"]["target"]
            error = tower_constr["mass"]["acceptable_error"]
            upper_value = target + error
            lower_value = target - error
            wt_opt.model.add_constraint("towerse.tower_mass", upper = upper_value, lower = lower_value, ref = 1.e+6)
            
        if monopile_constr["stress"]["flag"] and tower_constr["stress"]["flag"]:
            wt_opt.model.add_constraint("fixedse.post_monopile_tower.constr_stress", upper=1.0)
        elif monopile_constr["stress"]["flag"]:
            wt_opt.model.add_constraint("fixedse.post.constr_stress", upper=1.0)

        if monopile_constr["global_buckling"]["flag"] and tower_constr["global_buckling"]["flag"]:
            wt_opt.model.add_constraint("fixedse.post_monopile_tower.constr_global_buckling", upper=1.0)
        elif monopile_constr["global_buckling"]["flag"]:
            wt_opt.model.add_constraint("fixedse.post.constr_global_buckling", upper=1.0)

        if monopile_constr["shell_buckling"]["flag"] and tower_constr["shell_buckling"]["flag"]:
            wt_opt.model.add_constraint("fixedse.post_monopile_tower.constr_shell_buckling", upper=1.0)
        elif monopile_constr["shell_buckling"]["flag"]:
            wt_opt.model.add_constraint("fixedse.post.constr_shell_buckling", upper=1.0)

        if monopile_constr["d_to_t"]["flag"]:
            wt_opt.model.add_constraint(
                "fixedse.constr_d_to_t",
                lower=monopile_constr["d_to_t"]["lower_bound"],
                upper=monopile_constr["d_to_t"]["upper_bound"],
            )

        if monopile_constr["taper"]["flag"]:
            wt_opt.model.add_constraint("fixedse.constr_taper", lower=monopile_constr["taper"]["lower_bound"])

        if monopile_constr["slope"]["flag"]:
            wt_opt.model.add_constraint("fixedse.slope", upper=1.0)

        if monopile_constr["thickness_slope"]["flag"]:
            wt_opt.model.add_constraint("fixedse.thickness_slope", upper=1.0)

        if monopile_constr["frequency_1"]["flag"]:
            wt_opt.model.add_constraint(
                "fixedse.structural_frequencies",
                indices=[0],
                lower=monopile_constr["frequency_1"]["lower_bound"],
                upper=monopile_constr["frequency_1"]["upper_bound"],
            )

        if monopile_constr["pile_depth"]["flag"]:
            wt_opt.model.add_constraint("fixedse.suctionpile_depth", lower=monopile_constr["pile_depth"]["lower_bound"])

        if monopile_constr["tower_diameter_coupling"]["flag"]:
            wt_opt.model.add_constraint("fixedse.constr_diam_consistency", upper=1.0)
            
        if monopile_constr["mass"]["flag"]:
            target = monopile_constr["mass"]["target"]
            error = monopile_constr["mass"]["acceptable_error"]
            upper_value = target + error
            lower_value = target - error
            wt_opt.model.add_constraint("fixedse.monopile_mass", upper = upper_value, lower = lower_value, ref = 1.e+6)

        # Jacket constraints
        jacket_constr = self.opt["constraints"]["jacket"]

        if jacket_constr["stress"]["flag"]:
            wt_opt.model.add_constraint("fixedse.constr_stress", upper=1.0)

        if jacket_constr["global_buckling"]["flag"]:
            wt_opt.model.add_constraint("fixedse.constr_global_buckling", upper=1.0)

        if jacket_constr["shell_buckling"]["flag"]:
            wt_opt.model.add_constraint("fixedse.constr_shell_buckling", upper=1.0)

        if jacket_constr["tower_diameter_coupling"]["flag"]:
            wt_opt.model.add_constraint("fixedse.constr_diam_consistency", upper=1.0)

        if jacket_constr["frequency_1"]["flag"]:
            wt_opt.model.add_constraint(
                "fixedse.structural_frequencies",
                indices=[0],
                lower=jacket_constr["frequency_1"]["lower_bound"],
                upper=jacket_constr["frequency_1"]["upper_bound"],
            )

        # Hub and drivetrain constraints
        hub_constr = self.opt["constraints"]["hub"]
        drive_constr = self.opt["constraints"]["drivetrain"]

        if hub_constr["hub_diameter"]["flag"]:
            wt_opt.model.add_constraint("drivese.constr_hub_diameter", lower=0.0)

        for k in ["lss", "hss", "bedplate", "mb1", "mb2"]:
            if drive_constr[k]["flag"]:
                mystr = k + "_vonmises" if k in ["lss", "hss", "bedplate"] else k + "_defl"
                wt_opt.model.add_constraint("drivese.constr_" + mystr, upper=1.0)

        for k in ["shaft", "stator"]:
            for m in ["deflection", "angle"]:
                if drive_constr[f"{k}_{m}"]["flag"]:
                    wt_opt.model.add_constraint(f"drivese.constr_{k}_{m}", upper=1.0)

        for k in ["length", "height", "access", "ecc"]:
            if drive_constr[k]["flag"]:
                wt_opt.model.add_constraint("drivese.constr_" + k, lower=0.0)

        # Floating platform and mooring constraints
        float_constr = self.opt["constraints"]["floating"]

        if float_constr["buoyancy"]["flag"]:
            wt_opt.model.add_constraint("floatingse.variable_ballast_mass", lower=0.0)

        if float_constr["fixed_ballast_capacity"]["flag"]:
            wt_opt.model.add_constraint("floatingse.constr_fixed_margin", upper=1.0)

        if float_constr["variable_ballast_capacity"]["flag"]:
            wt_opt.model.add_constraint("floatingse.constr_variable_margin", lower=0.0, upper=1.0)

        if float_constr["metacentric_height"]["flag"]:
            wt_opt.model.add_constraint(
                "floatingse.metacentric_height_roll", lower=float_constr["metacentric_height"]["lower_bound"]
            )
            wt_opt.model.add_constraint(
                "floatingse.metacentric_height_pitch", lower=float_constr["metacentric_height"]["lower_bound"]
            )

        if float_constr["freeboard_margin"]["flag"]:
            wt_opt.model.add_constraint("floatingse.constr_freeboard_heel_margin", upper=1.0)

        if float_constr["draft_margin"]["flag"]:
            wt_opt.model.add_constraint("floatingse.constr_draft_heel_margin", upper=1.0)

        if float_constr["fairlead_depth"]["flag"]:
            wt_opt.model.add_constraint("floatingse.constr_fairlead_wave", upper=1.0)

        if float_constr["mooring_heel"]["flag"]:
            wt_opt.model.add_constraint("floatingse.constr_mooring_heel", upper=1.0)

        if float_constr["mooring_surge"]["flag"]:
            wt_opt.model.add_constraint("floatingse.constr_mooring_surge", upper=1.0)

        if float_constr["mooring_tension"]["flag"]:
            wt_opt.model.add_constraint("floatingse.constr_axial_load", upper=1.0)

        if float_constr["mooring_length"]["flag"]:
            wt_opt.model.add_constraint("floatingse.constr_mooring_length", upper=1.0)

        if float_constr["anchor_vertical"]["flag"]:
            wt_opt.model.add_constraint("floatingse.constr_anchor_vertical", lower=0.0)

        if float_constr["anchor_lateral"]["flag"]:
            wt_opt.model.add_constraint("floatingse.constr_anchor_lateral", lower=0.0)

        if float_constr["stress"]["flag"]:
            wt_opt.model.add_constraint("floatingse.constr_platform_stress", upper=1.0)

        if float_constr["shell_buckling"]["flag"]:
            wt_opt.model.add_constraint("floatingse.constr_platform_shell_buckling", upper=1.0)

        if float_constr["global_buckling"]["flag"]:
            wt_opt.model.add_constraint("floatingse.constr_platform_global_buckling", upper=1.0)

        for modestr in ["surge", "sway", "heave", "roll", "pitch", "yaw"]:
            if float_constr[f"{modestr}_period"]["flag"]:
                wt_opt.model.add_constraint(
                    f"{self.floating_period_solve_component}.{modestr}_period",
                    lower=float_constr[f"{modestr}_period"]["lower_bound"],
                    upper=float_constr[f"{modestr}_period"]["upper_bound"],
                )

        # User constraints
        user_constr = self.opt["constraints"]["user"]
        for k in range(len(user_constr)):
            var_k = user_constr[k]["name"]

            if "lower_bound" in user_constr[k]:
                lower_k = user_constr[k]["lower_bound"]
            elif "lower" in user_constr[k]:
                lower_k = user_constr[k]["lower"]
            else:
                lower_k = None

            if "upper_bound" in user_constr[k]:
                upper_k = user_constr[k]["upper_bound"]
            elif "upper" in user_constr[k]:
                upper_k = user_constr[k]["upper"]
            else:
                upper_k = None

            if "indices" in user_constr[k]:
                idx_k = user_constr[k]["indices"]
            else:
                idx_k = None
            
            if "ref" in user_constr[k]:
                ref_k = user_constr[k]["ref"]
            else:
                ref_k = None

            if lower_k is None and upper_k is None:
                raise Exception(f"Must include a lower_bound and/or an upper bound for {var_k}")

            wt_opt.model.add_constraint(var_k, lower=lower_k, upper=upper_k, indices=idx_k, ref=ref_k)

        return wt_opt

    def set_recorders(self, wt_opt):
        folder_output = self.opt["general"]["folder_output"]

        # Set recorder on the OpenMDAO driver level using the `optimization_log`
        # filename supplied in the optimization yaml
        if self.opt["recorder"]["flag"]:
            recorder = om.SqliteRecorder(os.path.join(folder_output, self.opt["recorder"]["file_name"]))
            wt_opt.driver.add_recorder(recorder)
            wt_opt.add_recorder(recorder)

            wt_opt.driver.recording_options["excludes"] = ["*_df"]
            wt_opt.driver.recording_options["record_constraints"] = True
            wt_opt.driver.recording_options["record_desvars"] = True
            wt_opt.driver.recording_options["record_objectives"] = True

            if self.opt["recorder"]["includes"]:
                wt_opt.driver.recording_options["includes"] = self.opt["recorder"]["includes"]

        return wt_opt

    def set_initial(self, wt_opt, wt_init):
        blade_opt = self.opt["design_variables"]["blade"]

        if self.modeling["flags"]["blade"]:
            wt_opt["blade.opt_var.s_opt_twist"] = np.linspace(0.0, 1.0, blade_opt["aero_shape"]["twist"]["n_opt"])
            twist_interpolator = PchipInterpolator(
                wt_init["components"]["blade"]["outer_shape_bem"]["twist"]["grid"],
                wt_init["components"]["blade"]["outer_shape_bem"]["twist"]["values"],
            )
            init_twist_opt = twist_interpolator(wt_opt["blade.opt_var.s_opt_twist"])
            wt_opt["blade.opt_var.twist_opt"] = init_twist_opt
            wt_opt["blade.opt_var.s_opt_chord"] = np.linspace(0.0, 1.0, blade_opt["aero_shape"]["chord"]["n_opt"])
            chord_interpolator = PchipInterpolator(
                wt_init["components"]["blade"]["outer_shape_bem"]["chord"]["grid"],
                wt_init["components"]["blade"]["outer_shape_bem"]["chord"]["values"],
            )
            init_chord_opt = chord_interpolator(wt_opt["blade.opt_var.s_opt_chord"])
            wt_opt["blade.opt_var.chord_opt"] = init_chord_opt
            if self.modeling["WISDEM"]["RotorSE"]["inn_af"]:
                wt_opt["inn_af.s_opt_r_thick"] = np.linspace(0.0, 1.0, blade_opt["aero_shape"]["rthick"]["n_opt"])
                r_thick_interpolator = PchipInterpolator(
                    wt_init["components"]["blade"]["outer_shape_bem"]["rthick"]["grid"],
                    wt_init["components"]["blade"]["outer_shape_bem"]["rthick"]["values"],
                )
                init_r_thick_opt = r_thick_interpolator(wt_opt["inn_af.s_opt_r_thick"])
                wt_opt["inn_af.r_thick_opt"] = init_r_thick_opt
                wt_opt["inn_af.s_opt_L_D"] = np.linspace(0.0, 1.0, blade_opt["aero_shape"]["L/D"]["n_opt"])
                L_D_interpolator = PchipInterpolator(
                    wt_init["components"]["blade"]["outer_shape_bem"]["L/D"]["grid"],
                    wt_init["components"]["blade"]["outer_shape_bem"]["L/D"]["values"],
                )
                init_L_D_opt = L_D_interpolator(wt_opt["inn_af.s_opt_L_D"])
                wt_opt["inn_af.L_D_opt"] = init_L_D_opt
                wt_opt["inn_af.s_opt_c_d"] = np.linspace(0.0, 1.0, blade_opt["aero_shape"]["c_d"]["n_opt"])
                c_d_interpolator = PchipInterpolator(
                    wt_init["components"]["blade"]["outer_shape_bem"]["c_d"]["grid"],
                    wt_init["components"]["blade"]["outer_shape_bem"]["c_d"]["values"],
                )
                init_c_d_opt = c_d_interpolator(wt_opt["inn_af.s_opt_c_d"])
                wt_opt["inn_af.c_d_opt"] = init_c_d_opt

                wt_opt["inn_af.s_opt_stall_margin"] = np.linspace(
                    0.0, 1.0, blade_opt["aero_shape"]["stall_margin"]["n_opt"]
                )
                stall_margin_interpolator = PchipInterpolator(
                    wt_init["components"]["blade"]["outer_shape_bem"]["stall_margin"]["grid"],
                    wt_init["components"]["blade"]["outer_shape_bem"]["stall_margin"]["values"],
                )
                init_stall_margin_opt = stall_margin_interpolator(wt_opt["inn_af.s_opt_stall_margin"])
                wt_opt["inn_af.stall_margin_opt"] = init_stall_margin_opt
            
            if not self.modeling["user_elastic"]["blade"]:
                # YL: no internal structure optimization when using user-defined blade elastic properties
                layers = wt_init["components"]["blade"]["internal_structure_2d_fem"]["layers"]
                for i in range(self.modeling["WISDEM"]["RotorSE"]["n_layers"]):
                    wt_opt["blade.opt_var.s_opt_layer_%d"%i] = np.linspace(
                        0.0, 1.0, blade_opt["n_opt_struct"][i]
                    )
                    thick_interp = PchipInterpolator(
                                layers[i]["thickness"]["grid"],
                                layers[i]["thickness"]["values"],
                                extrapolate=False)
                    init_opt = thick_interp(wt_opt["blade.opt_var.s_opt_layer_%d"%i])
                    wt_opt["blade.opt_var.layer_%d_opt"%i] = np.nan_to_num(init_opt, nan=0.)

                if self.modeling["WISDEM"]["RotorSE"]["flag"]:
                    blade_constr = self.opt["constraints"]["blade"]
                    wt_opt["rotorse.rs.constr.max_strainU_spar"] = blade_constr["strains_spar_cap_ss"]["max"]
                    wt_opt["rotorse.rs.constr.max_strainL_spar"] = blade_constr["strains_spar_cap_ps"]["max"]
                    if blade_constr["rail_transport"]["flag"]:
                        wt_opt["rotorse.re.rail.max_strains"] = min(
                            blade_constr["strains_spar_cap_ss"]["max"], blade_constr["strains_spar_cap_ps"]["max"]
                        )

                    wt_opt["rotorse.rs.constr.max_strainU_te"] = blade_constr["strains_te_ss"]["max"]
                    wt_opt["rotorse.rs.constr.max_strainL_te"] = blade_constr["strains_te_ps"]["max"]
                    wt_opt["rotorse.stall_check.stall_margin"] = blade_constr["stall"]["margin"] * 180.0 / np.pi
                    if self.modeling["flags"]["tower"]:
                        wt_opt["tcons.max_allowable_td_ratio"] = blade_constr["tip_deflection"]["margin"]

        if self.modeling["flags"]["nacelle"]:
            drive_constr = self.opt["constraints"]["drivetrain"]

            wt_opt["drivese.shaft_deflection_allowable"] = drive_constr["shaft_deflection"]["upper_bound"]
            wt_opt["drivese.shaft_angle_allowable"] = drive_constr["shaft_angle"]["upper_bound"]

            wt_opt["drivese.stator_deflection_allowable"] = drive_constr["stator_deflection"]["upper_bound"]
            wt_opt["drivese.stator_angle_allowable"] = drive_constr["stator_angle"]["upper_bound"]

            if self.modeling["WISDEM"]["DriveSE"]["direct"]:
                wt_opt["drivese.access_diameter"] = drive_constr["access"]["lower_bound"]

        if self.modeling["flags"]["floating"]:
            float_constr = self.opt["constraints"]["floating"]
            wt_opt["floatingse.max_surge_fraction"] = float_constr["max_surge"]["upper_bound"]
            wt_opt.set_val("floatingse.operational_heel", float_constr["operational_heel"]["upper_bound"], units="rad")
            wt_opt.set_val("floatingse.survival_heel", float_constr["survival_heel"]["upper_bound"], units="rad")

        return wt_opt

    def set_restart(self, wt_opt):
        if "warmstart_file" in self.opt["driver"]["optimization"]:
            # Directly read the pyoptsparse sqlite db file
            from pyoptsparse import SqliteDict

            db = SqliteDict(self.opt["driver"]["optimization"]["warmstart_file"])

            # Grab the last iteration's design variables
            last_key = db["last"]
            desvars = db[last_key]["xuser"]

            # Obtain the already-setup OM problem's design variables
            if wt_opt.model._static_mode:
                design_vars = wt_opt.model._static_design_vars
            else:
                design_vars = wt_opt.model._design_vars

            # Get the absolute names from the promoted names within the OM model.
            # We need this because the pyoptsparse db has the absolute names for
            # variables but the OM model uses the promoted names.
            prom2abs = wt_opt.model._var_allprocs_prom2abs_list["output"]
            abs2prom = {}
            for key in design_vars:
                abs2prom[prom2abs[key][0]] = key

            # Loop through each design variable
            for key in desvars:
                prom_key = abs2prom[key]

                # Scale each DV based on the OM scaling from the problem.
                # This assumes we're running the same problem with the same scaling
                scaler = design_vars[prom_key]["scaler"]
                adder = design_vars[prom_key]["adder"]

                if scaler is None:
                    scaler = 1.0
                if adder is None:
                    adder = 0.0

                scaled_dv = desvars[key] / scaler - adder

                # Special handling for blade twist as we only have the
                # last few control points as design variables
                if "twist_opt" in key:
                    wt_opt[key][2:] = scaled_dv
                else:
                    wt_opt[key][:] = scaled_dv

        return wt_opt



# Import the libraries we need
import numpy as np
import openmdao.api as om

# -------


# Create the mock mathematical relationships associated with Discipline 1 using an OpenMDAO component
class SellarDis1(om.ExplicitComponent):
    """
    Component containing Discipline 1 -- no derivatives version.
    """

    def setup(self):
        # Global Design Variable
        self.add_input("z", val=np.zeros(2))

        # Local Design Variable
        self.add_input("x", val=0.0)

        # Coupling parameter
        self.add_input("y2", val=1.0)

        # Coupling output
        self.add_output("y1", val=1.0)

        # Finite difference all partials.
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs):
        """
        Evaluates the equation
        y1 = z1**2 + z2 + x1 - 0.2*y2
        """
        z1 = inputs["z"][0]
        z2 = inputs["z"][1]
        x1 = inputs["x"]
        y2 = inputs["y2"]

        outputs["y1"] = z1**2 + z2 + x1 - 0.2 * y2
        # -- end Discipline 1


# Now create the mock mathematical relationships associated with Discipline 2 using an OpenMDAO component
class SellarDis2(om.ExplicitComponent):
    """
    Component containing Discipline 2 -- no derivatives version.
    """

    def setup(self):
        # Global Design Variable
        self.add_input("z", val=np.zeros(2))

        # Coupling parameter
        self.add_input("y1", val=1.0)

        # Coupling output
        self.add_output("y2", val=1.0)

        # Finite difference all partials.
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs):
        """
        Evaluates the equation
        y2 = y1**(.5) + z1 + z2
        """

        z1 = inputs["z"][0]
        z2 = inputs["z"][1]
        y1 = inputs["y1"]

        outputs["y2"] = y1**0.5 + z1 + z2
        # -- end Discipline 2


# Assemble the system together in an OpenMDAO Group
class SellarMDA(om.Group):
    """
    Group containing the Sellar MDA.
    """

    def setup(self):
        indeps = self.add_subsystem("indeps", om.IndepVarComp(), promotes=["*"])
        indeps.add_output("x", 1.0)
        indeps.add_output("z", np.array([5.0, 2.0]))

        self.add_subsystem("d1", SellarDis1(), promotes=["y1", "y2"])
        self.add_subsystem("d2", SellarDis2(), promotes=["y1", "y2"])
        self.connect("x", "d1.x")
        self.connect("z", ["d1.z", "d2.z"])

        # Nonlinear Block Gauss Seidel is a gradient free solver to handle implicit loops
        self.nonlinear_solver = om.NonlinearBlockGS()

        self.add_subsystem(
            "obj_cmp",
            om.ExecComp("obj = x**2 + z[1] + y1 + exp(-y2)", z=np.array([0.0, 0.0]), x=0.0),
            promotes=["x", "z", "y1", "y2", "obj"],
        )

        self.add_subsystem("con_cmp1", om.ExecComp("con1 = 3.16 - y1"), promotes=["con1", "y1"])
        self.add_subsystem("con_cmp2", om.ExecComp("con2 = y2 - 24.0"), promotes=["con2", "y2"])
        # -- end Group


# Instantiate the model
prob = om.Problem()
prob.model = SellarMDA()
# --------

# Specify the optimization 'driver' and options
prob.driver = om.ScipyOptimizeDriver()
prob.driver.options["optimizer"] = "SLSQP"
prob.driver.options["tol"] = 1e-8
# --------

# Assign objective and design variables
prob.model.add_design_var("x", lower=0, upper=10)
prob.model.add_design_var("z", lower=0, upper=10)
prob.model.add_objective("obj")
prob.model.add_constraint("con1", upper=0)
prob.model.add_constraint("con2", upper=0)

# Ask OpenMDAO to finite-difference across the whole model to compute the total gradients for the optimizer
# The other approach would be to finite-difference for the partials and build up the total derivative
prob.model.approx_totals()
# --------

# Execute! and display result
prob.setup()
prob.run_driver()

print("minimum found at")
print(float(prob["x"]))
print(prob["z"])
print("minimum objective")
print(float(prob["obj"]))
# --------



from wisdem.precomp.precomp import PreComp
from wisdem.precomp.profile import Profile
from wisdem.precomp.composite_section import CompositeSection
from wisdem.precomp.orthotropic import Orthotropic2DMaterial




import unittest

import numpy as np
from openmdao.utils.assert_utils import assert_near_equal

from wisdem.optimization_drivers.dakota_driver import DakotaOptimizer

try:
    import dakota
except ImportError:
    dakota = None


@unittest.skipIf(dakota is None, "only run if Dakota is installed.")
class TestDakotaOptimization(unittest.TestCase):
    def test_2D_opt_max_iterations(self):
        bounds = {"x": np.array([[0.0, 1.0], [0.0, 1.0]])}
        desvars = {"x": np.array([0.0, 0.25])}
        outputs = ["y"]
        template_dir = "template_dir/"
        model_string = "from weis.multifidelity.models.testbed_components import simple_2D_high_model as model"
        output_scalers = [1.0]
        options = {"method": "coliny_cobyla", "max_function_evaluations": 3}

        opt = DakotaOptimizer(template_dir)
        results = opt.optimize(desvars, outputs, bounds, model_string, output_scalers, options)

        assert_near_equal(np.min(np.array(results["y"])), -9.5)

    def test_2D_opt_EGO(self):
        bounds = {"x": np.array([[0.0, 1.0], [0.0, 1.0]])}
        desvars = {"x": np.array([0.0, 0.25])}
        outputs = ["y"]
        template_dir = "template_dir/"
        model_string = "from weis.multifidelity.models.testbed_components import simple_2D_high_model as model"
        output_scalers = [1.0]
        options = {"initial_samples": 5, "method": "efficient_global", "seed": 123456}

        opt = DakotaOptimizer(template_dir)
        results = opt.optimize(desvars, outputs, bounds, model_string, output_scalers, options)

        assert_near_equal(np.min(np.array(results["y"])), -9.999996864)

    def test_two_variables(self):
        bounds = {"x": np.array([[0.0, 1.0], [0.0, 1.0]]), "z": [1.0, 2.0]}
        desvars = {"x": np.array([0.0, 0.25]), "z": 1.5}
        outputs = ["y"]
        template_dir = "template_dir/"
        model_string = "from weis.multifidelity.models.testbed_components import simple_two_variable as model"
        output_scalers = [1.0]
        options = {"method": "coliny_cobyla", "max_function_evaluations": 3}

        opt = DakotaOptimizer(template_dir)
        results = opt.optimize(desvars, outputs, bounds, model_string, output_scalers, options)

        assert_near_equal(np.min(np.array(results["y"])), 1.0)

    def test_constraint(self):
        bounds = {"x": np.array([[0.0, 1.0], [0.0, 1.0]])}
        desvars = {"x": np.array([0.0, 0.25])}
        outputs = ["y", "con"]
        template_dir = "template_dir/"
        model_string = "from weis.multifidelity.models.testbed_components import simple_2D_low_model as model"
        output_scalers = [1.0, 1.0]
        options = {"method": "coliny_cobyla", "max_function_evaluations": 3}

        opt = DakotaOptimizer(template_dir)
        results = opt.optimize(desvars, outputs, bounds, model_string, output_scalers, options)

        assert_near_equal(np.min(np.array(results["y"])), 0.5)
        assert_near_equal(np.min(np.array(results["con"])), 0.0)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# encoding: utf-8
"""
precomp.py

Created by Andrew Ning on 2012-02-28.
Copyright (c)  NREL. All rights reserved.
"""

import os
import numpy as np

#from wisdem.precomp._precomp import precomp as _precomp
import wisdem.precomp.properties as properties


def web_loc(r, chord, le, ib_idx, ob_idx, ib_webc, ob_webc):
    n = len(r)
    loc = np.zeros(n)

    for i in range(n):
        if i < ib_idx or i > ob_idx:
            loc[i] = -1
        else:
            xn = (r[i] - r[ib_idx]) / (r[ob_idx] - r[ib_idx])
            loc[i] = (
                le[i]
                - (le[ib_idx] - ib_webc) * chord[ib_idx] / chord[i] * (1 - xn)
                - (le[ob_idx] - ob_webc) * chord[ob_idx] / chord[i] * xn
            )

    return loc


class PreComp:
    def __init__(
        self,
        r,
        chord,
        theta,
        leLoc,
        precurve,
        presweep,
        profile,
        materials,
        upperCS,
        lowerCS,
        websCS,
        sector_idx_strain_spar_ps,
        sector_idx_strain_spar_ss,
        sector_idx_strain_te_ps,
        sector_idx_strain_te_ss,
    ):
        """Constructor

        Parameters
        ----------
        r : ndarray (m)
            radial positions. r[0] should be the hub location
            while r[-1] should be the blade tip. Any number
            of locations can be specified between these in ascending order.
        chord : ndarray (m)
            array of chord lengths at corresponding radial positions
        theta : ndarray (deg)
            array of twist angles at corresponding radial positions.
            (positive twist decreases angle of attack)
        leLoc : ndarray(float)
            array of leading-edge positions from a reference blade axis (usually blade pitch axis).
            locations are normalized by the local chord length.  e.g. leLoc[i] = 0.2 means leading edge
            is 0.2*chord[i] from reference axis.   positive in -x direction for airfoil-aligned coordinate system
        profile : list(:class:`Profile`)
            airfoil shape at each radial position
        materials : list(:class:`Orthotropic2DMaterial`), optional
            list of all Orthotropic2DMaterial objects used in defining the geometry
        upperCS, lowerCS, websCS : list(:class:`CompositeSection`)
            list of CompositeSection objections defining the properties for upper surface, lower surface,
            and shear webs (if any) for each section

        """

        self.r = np.array(r)
        self.chord = np.array(chord)
        self.theta = np.array(theta)
        self.leLoc = np.array(leLoc)
        self.precurve = np.array(precurve)
        self.presweep = np.array(presweep)

        self.profile = profile
        self.materials = materials
        self.upperCS = upperCS
        self.lowerCS = lowerCS
        self.websCS = websCS

        self.sector_idx_strain_spar_ps = sector_idx_strain_spar_ps
        self.sector_idx_strain_spar_ss = sector_idx_strain_spar_ss
        self.sector_idx_strain_te_ps = sector_idx_strain_te_ps
        self.sector_idx_strain_te_ss = sector_idx_strain_te_ss

        # twist rate
        #self.th_prime = _precomp.tw_rate(self.r, self.theta)
        self.th_prime = properties.tw_rate(self.r, self.theta)

    def sectionProperties(self):
        """see meth:`SectionStrucInterface.sectionProperties`"""

        # radial discretization
        nsec = len(self.r)

        # initialize variables
        beam_z = self.r
        beam_EA = np.zeros(nsec)
        beam_EIxx = np.zeros(nsec)
        beam_EIyy = np.zeros(nsec)
        beam_EIxy = np.zeros(nsec)
        beam_GJ = np.zeros(nsec)
        beam_EA_EIxx = np.zeros(nsec)
        beam_EA_EIyy = np.zeros(nsec)
        beam_EIxx_GJ = np.zeros(nsec)
        beam_EIyy_GJ = np.zeros(nsec)
        beam_EA_GJ = np.zeros(nsec)
        beam_rhoA = np.zeros(nsec)
        beam_A = np.zeros(nsec)
        beam_rhoJ = np.zeros(nsec)
        beam_Tw_iner = np.zeros(nsec)

        beam_flap_iner = np.zeros(nsec)
        beam_edge_iner = np.zeros(nsec)

        beam_x_sc = np.zeros(nsec)
        beam_y_sc = np.zeros(nsec)
        beam_x_tc = np.zeros(nsec)
        beam_y_tc = np.zeros(nsec)
        beam_x_cg = np.zeros(nsec)
        beam_y_cg = np.zeros(nsec)

        # distance to elastic center from point about which structural properties are computed
        # using airfoil coordinate system
        beam_x_ec = np.zeros(nsec)
        beam_y_ec = np.zeros(nsec)

        # distance to elastic center from airfoil nose
        # using profile coordinate system
        self.x_ec_nose = np.zeros(nsec)
        self.y_ec_nose = np.zeros(nsec)

        profile = self.profile
        mat = self.materials
        csU = self.upperCS
        csL = self.lowerCS
        csW = self.websCS

        # arrange materials into array
        n = len(mat)
        E1 = [0] * n
        E2 = [0] * n
        G12 = [0] * n
        nu12 = [0] * n
        rho = [0] * n

        for i in range(n):
            E1[i] = mat[i].E1
            E2[i] = mat[i].E2
            G12[i] = mat[i].G12
            nu12[i] = mat[i].nu12
            rho[i] = mat[i].rho

        for i in range(nsec):
            # print(i)

            xnode, ynode = profile[i]._preCompFormat()
            locU, n_laminaU, n_pliesU, tU, thetaU, mat_idxU = csU[i]._preCompFormat()
            locL, n_laminaL, n_pliesL, tL, thetaL, mat_idxL = csL[i]._preCompFormat()
            locW, n_laminaW, n_pliesW, tW, thetaW, mat_idxW = csW[i]._preCompFormat()

            nwebs = len(locW)

            # address a bug in f2py (need to pass in length 1 arrays even though they are not used)
            if nwebs == 0:
                locW = [0]
                n_laminaW = [0]
                n_pliesW = [0]
                tW = [0]
                thetaW = [0]
                mat_idxW = [0]

            #(eifbar,eilbar,gjbar,eabar,eiflbar,
            # sfbar,slbar,sftbar,sltbar,satbar,
            # z_sc,y_sc,ztc_ref,ytc_ref,
            # mass,area,iflap_eta,ilag_zeta,tw_iner,
            # zcm_ref,ycm_ref) = _precomp.properties(
            #    self.chord[i],
            #    self.theta[i],
            #    self.th_prime[i],
            #    self.leLoc[i],
            #    xnode,
            #    ynode,
            #    E1,
            #    E2,
            #    G12,
            #    nu12,
            #    rho,
            #    locU,
            #    n_laminaU,
            #    n_pliesU,
            #    tU,
            #    thetaU,
            #    mat_idxU,
            #    locL,
            #    n_laminaL,
            #    n_pliesL,
            #    tL,
            #    thetaL,
            #    mat_idxL,
            #    nwebs,
            #    locW,
            #    n_laminaW,
            #    n_pliesW,
            #    tW,
            #    thetaW,
            #    mat_idxW,
            #)

            (eifbar,eilbar,gjbar,eabar,eiflbar,
             sfbar,slbar,sftbar,sltbar,satbar,
             z_sc,y_sc,ztc_ref,ytc_ref,
             mass,area,iflap_eta,ilag_zeta,tw_iner,
             zcm_ref,ycm_ref) = properties.properties(
                self.chord[i],
                self.theta[i],
                self.th_prime[i],
                self.leLoc[i],
                xnode,
                ynode,
                E1,
                E2,
                G12,
                nu12,
                rho,
                locU,
                n_laminaU,
                n_pliesU,
                tU,
                thetaU,
                mat_idxU,
                locL,
                n_laminaL,
                n_pliesL,
                tL,
                thetaL,
                mat_idxL,
                nwebs,
                locW,
                n_laminaW,
                n_pliesW,
                tW,
                thetaW,
                mat_idxW,
            )
            
            beam_EIxx[i] = eilbar  # EI_lag, Section lag (edgewise) bending stiffness about the XE axis (Nm2)
            beam_EIyy[i] = eifbar  # EI_flap, Section flap bending stiffness about the YE axis (Nm2)
            beam_GJ[i] = gjbar #  Section torsion stiffness (Nm2)
            beam_EA[i] = eabar # Section axial stiffness (N)
            beam_EIxy[i] = eiflbar # Coupled flap-lag stiffness with respect to the XE-YE frame (Nm2)
            beam_EA_EIxx[i] = slbar # Coupled axial-lag stiffness with respect to the XE-YE frame (Nm.)
            beam_EA_EIyy[i] = sfbar # Coupled axial-flap stiffness with respect to the XE-YE frame (Nm)
            beam_EIxx_GJ[i] = sltbar # Coupled lag-torsion stiffness with respect to the XE-YE frame (Nm2)
            beam_EIyy_GJ[i] = sftbar # Coupled flap-torsion stiffness with respect to the XE-YE frame (Nm2)
            beam_EA_GJ[i] = satbar # Coupled axial-torsion stiffness (Nm)
            beam_x_sc[i] = z_sc # X-coordinate of the shear-center offset with respect to the ref axes (m)
            beam_y_sc[i] = y_sc # Chordwise offset of the section shear-center with respect to the reference frame, XR-YR (m)
            beam_x_tc[i] = ztc_ref # X-coordinate of the tension-center offset with respect to the XR-YR axes (m)
            beam_y_tc[i] = ytc_ref # Chordwise offset of the section tension-center with respect to the XR-YR axes (m)
            beam_rhoA[i] = mass # Section mass per unit length (kg/m)
            beam_A[i] = area # Cross-Sectional area (m)
            beam_flap_iner[i] = iflap_eta # Section flap inertia about the YG axis per unit length (kg-m)
            beam_edge_iner[i] = ilag_zeta # Section lag inertia about the XG axis per unit length (kg-m)
            beam_Tw_iner[i] = tw_iner # Orientation of the section principal inertia axes with respect the blade reference plane, theta (deg)
            beam_x_cg[i] = zcm_ref # X-coordinate of the center-of-mass offset with respect to the XR-YR axes (m)
            beam_y_cg[i] = ycm_ref # Chordwise offset of the section center of mass with respect to the XR-YR axes (m)


        beam_rhoJ = beam_flap_iner + beam_edge_iner  # perpendicular axis theorem

        self.x_ec_nose = beam_y_tc + self.leLoc * self.chord
        self.y_ec_nose = beam_x_tc  # switch b.c of coordinate system used
            
        return (
            beam_EIxx,
            beam_EIyy,
            beam_GJ,
            beam_EA,
            beam_EIxy,
            beam_EA_EIxx,
            beam_EA_EIyy,
            beam_EIxx_GJ,
            beam_EIyy_GJ,
            beam_EA_GJ,
            beam_rhoA,
            beam_A,
            beam_rhoJ,
            beam_Tw_iner,
            beam_flap_iner,
            beam_edge_iner,
            beam_x_tc,
            beam_y_tc,
            beam_x_sc,
            beam_y_sc,
            beam_x_cg,
            beam_y_cg,
        )

    def criticalStrainLocations(self, sector_idx_strain_ss, sector_idx_strain_ps):
        n = len(self.r)

        # find location of max thickness on airfoil
        xun = np.zeros(n)
        xln = np.zeros(n)
        yun = np.zeros(n)
        yln = np.zeros(n)

        # for i, p in enumerate(self.profile):
        #     xun[i], yun[i], yln[i] = p.locationOfMaxThickness()
        # xln = xun

        for i in range(n):
            csU = self.upperCS[i]
            csL = self.lowerCS[i]
            pf = self.profile[i]
            idx_ss = sector_idx_strain_ss[i]
            idx_ps = sector_idx_strain_ps[i]

            if idx_ps is None:
                xln[i] = 0.0
                yln[i] = 0.0
            else:
                xln[i] = 0.5 * (csL.loc[idx_ps] + csL.loc[idx_ps + 1])
                yln[i] = np.interp(xln[i], pf.x, pf.yl)

            if idx_ss is None:
                xun[i] = 0.0
                yun[i] = 0.0
            else:
                xun[i] = 0.5 * (csU.loc[idx_ss] + csU.loc[idx_ss + 1])
                yun[i] = np.interp(xun[i], pf.x, pf.yu)

        # make dimensional and define relative to elastic center
        xu = xun * self.chord - self.x_ec_nose
        xl = xln * self.chord - self.x_ec_nose
        yu = yun * self.chord - self.y_ec_nose
        yl = yln * self.chord - self.y_ec_nose

        # switch to airfoil coordinate system
        xu, yu = yu, xu
        xl, yl = yl, xl

        return xu, xl, yu, yl

    def panelBucklingStrain(self, sector_idx_array):
        """
        see chapter on Structural Component Design Techniques from Alastair Johnson
        section 6.2: Design of composite panels

        assumes: large aspect ratio, simply supported, uniaxial compression, flat rectangular plate

        """
        chord = self.chord
        nsec = len(self.r)

        eps_crit = np.zeros(nsec)

        for i in range(nsec):
            cs = self.upperCS[i]  # TODO: lower surface may be the compression one
            sector_idx = sector_idx_array[i]

            if sector_idx is None:
                eps_crit[i] = 0.0

            else:
                # chord-wise length of sector
                sector_length = chord[i] * (cs.loc[sector_idx + 1] - cs.loc[sector_idx])

                # get matrices
                A, B, D, totalHeight = cs.compositeMatrices(sector_idx)
                E = cs.effectiveEAxial(sector_idx)
                D1 = D[0, 0]
                D2 = D[1, 1]
                D3 = D[0, 1] + 2 * D[2, 2]

                # use empirical formula
                Nxx = 2 * (np.pi / sector_length) ** 2 * (np.sqrt(D1 * D2) + D3)
                # Nxx = 3.6 * (np.pi/sector_length)**2 * D1

                eps_crit[i] = -Nxx / totalHeight / E

        return eps_crit


if __name__ == "__main__":
    import os

    # geometry
    r_str = [
        1.5,
        1.80135,
        1.89975,
        1.99815,
        2.1027,
        2.2011,
        2.2995,
        2.87145,
        3.0006,
        3.099,
        5.60205,
        6.9981,
        8.33265,
        10.49745,
        11.75205,
        13.49865,
        15.84795,
        18.4986,
        19.95,
        21.99795,
        24.05205,
        26.1,
        28.14795,
        32.25,
        33.49845,
        36.35205,
        38.4984,
        40.44795,
        42.50205,
        43.49835,
        44.55,
        46.49955,
        48.65205,
        52.74795,
        56.16735,
        58.89795,
        61.62855,
        63.0,
    ]
    chord_str = [
        3.386,
        3.386,
        3.386,
        3.386,
        3.386,
        3.386,
        3.386,
        3.386,
        3.387,
        3.39,
        3.741,
        4.035,
        4.25,
        4.478,
        4.557,
        4.616,
        4.652,
        4.543,
        4.458,
        4.356,
        4.249,
        4.131,
        4.007,
        3.748,
        3.672,
        3.502,
        3.373,
        3.256,
        3.133,
        3.073,
        3.01,
        2.893,
        2.764,
        2.518,
        2.313,
        2.086,
        1.419,
        1.085,
    ]
    theta_str = [
        13.31,
        13.31,
        13.31,
        13.31,
        13.31,
        13.31,
        13.31,
        13.31,
        13.31,
        13.31,
        13.31,
        13.31,
        13.31,
        13.31,
        13.31,
        12.53,
        11.48,
        10.63,
        10.16,
        9.59,
        9.01,
        8.4,
        7.79,
        6.54,
        6.18,
        5.36,
        4.75,
        4.19,
        3.66,
        3.4,
        3.13,
        2.74,
        2.32,
        1.53,
        0.86,
        0.37,
        0.11,
        0.0,
    ]
    le_str = [
        0.5,
        0.5,
        0.5,
        0.5,
        0.5,
        0.5,
        0.5,
        0.5,
        0.498,
        0.497,
        0.465,
        0.447,
        0.43,
        0.411,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
        0.4,
    ]
    web1 = np.array(
        [
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            0.4114,
            0.4102,
            0.4094,
            0.3876,
            0.3755,
            0.3639,
            0.345,
            0.3342,
            0.3313,
            0.3274,
            0.323,
            0.3206,
            0.3172,
            0.3138,
            0.3104,
            0.307,
            0.3003,
            0.2982,
            0.2935,
            0.2899,
            0.2867,
            0.2833,
            0.2817,
            0.2799,
            0.2767,
            0.2731,
            0.2664,
            0.2607,
            0.2562,
            0.1886,
            -1.0,
        ]
    )
    web2 = np.array(
        [
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            0.5886,
            0.5868,
            0.5854,
            0.5508,
            0.5315,
            0.5131,
            0.4831,
            0.4658,
            0.4687,
            0.4726,
            0.477,
            0.4794,
            0.4828,
            0.4862,
            0.4896,
            0.493,
            0.4997,
            0.5018,
            0.5065,
            0.5101,
            0.5133,
            0.5167,
            0.5183,
            0.5201,
            0.5233,
            0.5269,
            0.5336,
            0.5393,
            0.5438,
            0.6114,
            -1.0,
        ]
    )
    web3 = np.array(
        [
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
        ]
    )
    precurve_str = np.zeros_like(r_str)
    presweep_str = np.zeros_like(r_str)

    # -------- materials and composite layup  -----------------
    basepath = os.path.join("5MW_files", "5MW_PreCompFiles")

    materials = Orthotropic2DMaterial.listFromPreCompFile(os.path.join(basepath, "materials.inp"))

    ncomp = len(r_str)
    upper = [0] * ncomp
    lower = [0] * ncomp
    webs = [0] * ncomp
    profile = [0] * ncomp

    # # web 1
    # ib_idx = 7
    # ob_idx = 36
    # ib_webc = 0.4114
    # ob_webc = 0.1886

    # web1 = web_loc(r_str, chord_str, le_str, ib_idx, ob_idx, ib_webc, ob_webc)

    # # web 2
    # ib_idx = 7
    # ob_idx = 36
    # ib_webc = 0.5886
    # ob_webc = 0.6114

    # web2 = web_loc(r_str, chord_str, le_str, ib_idx, ob_idx, ib_webc, ob_webc)

    for i in range(ncomp):
        webLoc = []
        if web1[i] != -1:
            webLoc.append(web1[i])
        if web2[i] != -1:
            webLoc.append(web2[i])
        if web3[i] != -1:
            webLoc.append(web3[i])

        upper[i], lower[i], webs[i] = CompositeSection.initFromPreCompLayupFile(
            os.path.join(basepath, "layup_" + str(i + 1) + ".inp"), webLoc, materials
        )
        profile[i] = Profile.initFromPreCompFile(os.path.join(basepath, "shape_" + str(i + 1) + ".inp"))
    # --------------------------------------

    precomp = PreComp(
        r_str, chord_str, theta_str, le_str, precurve_str, presweep_str, profile, materials, upper, lower, webs
    )

    # evalute section properties
    EA, EIxx, EIyy, EIxy, GJ, rhoA, rhoJ, x_ec_str, y_ec_str = precomp.sectionProperties()

    import matplotlib.pyplot as plt

    r_str = np.array(r_str)
    rstar = (r_str - r_str[0]) / (r_str[-1] - r_str[0])

    plt.figure(1)
    plt.semilogy(rstar, EIxx)
    plt.xlabel("blade fraction")
    plt.ylabel("Edgewise Stiffness ($N m^2$)")

    plt.figure(2)
    plt.semilogy(rstar, EIyy)
    plt.xlabel("blade fraction")
    plt.ylabel("Flapwise Stiffness ($N m^2$)")

    plt.figure(3)
    plt.semilogy(rstar, EA)
    plt.figure(4)
    plt.semilogy(rstar, EIxy)
    plt.figure(5)
    plt.semilogy(rstar, GJ)
    plt.figure(6)
    plt.semilogy(rstar, rhoA)
    plt.figure(7)
    plt.semilogy(rstar, rhoJ)
    plt.figure(8)
    plt.plot(rstar, x_ec_str)
    plt.figure(9)
    plt.plot(rstar, y_ec_str)
    plt.figure(10)
    plt.plot(rstar, precomp.x_ec_nose)
    plt.figure(11)
    plt.plot(rstar, precomp.y_ec_nose)

    plt.show()



import os
import sys
import time

from openmdao.utils.mpi import MPI
from wisdem.glue_code.runWISDEM import run_wisdem

## File management
run_dir = os.path.dirname(os.path.realpath(__file__))
fname_wt_input_oc3 = run_dir + os.sep + "nrel5mw-spar_oc3_user_mass.yaml"
fname_modeling_options = run_dir + os.sep + "modeling_options.yaml"
fname_analysis_options = run_dir + os.sep + "analysis_options.yaml"


tt = time.time()
wt_opt, modeling_options, opt_options = run_wisdem(fname_wt_input_oc3, fname_modeling_options, fname_analysis_options)

if MPI:
    rank = MPI.COMM_WORLD.Get_rank()
else:
    rank = 0
if rank == 0:
    print("Run time: %f" % (time.time() - tt))
    sys.stdout.flush()



#!/usr/bin/env python
# encoding: utf-8

import numpy as np
import openmdao.api as om
from scipy.special import ellipeinc

import wisdem.commonse.utilities as util
from wisdem.commonse.cross_sections import IBeam


def rod_prop(s, Di, ti, rho):
    L = s.max() - s.min()

    def equal_pts(xi):
        if len(xi) < len(s) and len(xi) == 2:
            x = np.interp((s - s.min()) / L, [0, 1], xi)
        elif len(xi) == len(s):
            x = xi
        else:
            raise ValueError("Unknown grid of input", str(xi))
        return x

    D = equal_pts(Di)
    t = equal_pts(ti)
    y = 0.25 * rho * np.pi * (D**2 - (D - 2 * t) ** 2)
    m = np.trapz(y, s)
    cm = np.trapz(y * s, s) / m
    Dm = D.mean()
    tm = t.mean()
    I = np.array(
        [
            0.5 * 0.25 * (Dm**2 + (Dm - 2 * tm) ** 2),
            (1.0 / 12.0) * (3 * 0.25 * (Dm**2 + (Dm - 2 * tm) ** 2) + L**2),
            (1.0 / 12.0) * (3 * 0.25 * (Dm**2 + (Dm - 2 * tm) ** 2) + L**2),
        ]
    )
    return m, cm, m * I


class Layout(om.ExplicitComponent):
    """
    Calculate lengths, heights, and diameters of key drivetrain components in a
    direct drive system (valid for upwind or downwind).

    Parameters
    ----------
    upwind : boolean
        Flag whether the design is upwind or downwind
    L_12 : float, [m]
        Length from bearing #1 to bearing #2
    L_h1 : float, [m]
        Length from hub / start of lss to bearing #1
    L_generator : float, [m]
        Generator stack width
    overhang : float, [m]
        Overhang of rotor from tower along x-axis in yaw-aligned c.s.
    drive_height : float, [m]
        Hub height above tower top
    tilt : float, [deg]
        Angle of drivetrain lss tilt
    lss_diameter : numpy array[2], [m]
        LSS outer diameter from hub to bearing 2
    lss_wall_thickness : numpy array[2], [m]
        LSS wall thickness
    hub_diameter : float, [m]
        Diameter of hub
    D_top : float, [m]
        Tower top outer diameter
    lss_rho : float, [kg/m**3]
        material density
    bedplate_rho : float, [kg/m**3]
        material density
    bedplate_mass_user : float, [kg]
        User override of bedplate mass

    Returns
    -------
    L_lss : float, [m]
        Length of nose
    L_drive : float, [m]
        Length of drivetrain from bedplate to hub flang
    s_lss : numpy array[5], [m]
        LSS discretized s-coordinates
    lss_mass : float, [kg]
        LSS mass
    lss_cm : float, [m]
        LSS center of mass along lss axis from bedplate
    lss_I : numpy array[3], [kg*m**2]
        LSS moment of inertia around cm in axial (hub-aligned) c.s.
    L_bedplate : float, [m]
        Length of bedplate
    H_bedplate : float, [m]
        height of bedplate
    bedplate_mass : float, [kg]
        Bedplate mass
    bedplate_cm : numpy array[3], [m]
        Bedplate center of mass
    bedplate_I : numpy array[6], [kg*m**2]
        Bedplate mass moment of inertia about base
    s_mb1 : float, [m]
        Bearing 1 s-coordinate along drivetrain, measured from bedplate
    s_mb2 : float, [m]
        Bearing 2 s-coordinate along drivetrain, measured from bedplate
    s_gearbox : float, [m]
        Overall gearbox cm
    s_generator : float, [m]
        Overall generator cm
    constr_length : float, [m]
        Margin for drivetrain length and desired overhang distance (should be > 0)
    constr_height : float, [m]
        Margin for drivetrain height and desired hub height (should be > 0)

    """

    def setup(self):
        self.add_discrete_input("upwind", True)
        self.add_input("L_12", 0.0, units="m")
        self.add_input("L_h1", 0.0, units="m")
        self.add_input("L_generator", 0.0, units="m")
        self.add_input("overhang", 0.0, units="m")
        self.add_input("drive_height", 0.0, units="m")
        self.add_input("tilt", 0.0, units="deg")
        self.add_input("lss_diameter", np.zeros(2), units="m")
        self.add_input("lss_wall_thickness", np.zeros(2), units="m")
        self.add_input("D_top", 0.0, units="m")
        self.add_input("hub_diameter", val=0.0, units="m")
        self.add_input("lss_rho", val=0.0, units="kg/m**3")
        self.add_input("bedplate_rho", val=0.0, units="kg/m**3")
        self.add_input("bedplate_mass_user", val=0.0, units="kg")

        self.add_output("L_lss", 0.0, units="m")
        self.add_output("L_drive", 0.0, units="m")
        self.add_output("s_lss", val=np.zeros(5), units="m")
        self.add_output("lss_mass", val=0.0, units="kg")
        self.add_output("lss_cm", val=0.0, units="m")
        self.add_output("lss_I", val=np.zeros(3), units="kg*m**2")
        self.add_output("L_bedplate", 0.0, units="m")
        self.add_output("H_bedplate", 0.0, units="m")
        self.add_output("bedplate_mass", val=0.0, units="kg")
        self.add_output("bedplate_cm", val=np.zeros(3), units="m")
        self.add_output("bedplate_I", val=np.zeros(6), units="kg*m**2")
        self.add_output("s_mb1", val=0.0, units="m")
        self.add_output("s_mb2", val=0.0, units="m")
        self.add_output("s_stator", val=0.0, units="m")
        self.add_output("s_rotor", val=0.0, units="m")
        self.add_output("s_gearbox", val=0.0, units="m")
        self.add_output("s_generator", val=0.0, units="m")
        self.add_output("hss_mass", val=0.0, units="kg")
        self.add_output("hss_cm", val=0.0, units="m")
        self.add_output("hss_I", val=np.zeros(3), units="kg*m**2")
        self.add_output("constr_length", 0.0, units="m")
        self.add_output("constr_height", 0.0, units="m")


class DirectLayout(Layout):
    """
    Calculate lengths, heights, and diameters of key drivetrain components in a
    direct drive system (valid for upwind or downwind).

    Parameters
    ----------
    access_diameter : float, [m]
        Minimum diameter required for maintenance access
    nose_diameter : numpy array[2], [m]
        Nose outer diameter from bearing 1 to bedplate
    nose_wall_thickness : numpy array[2], [m]
        Nose wall thickness
    bedplate_wall_thickness : numpy array[4], [m]
        Bedplate wall thickness

    Returns
    -------
    L_nose : float, [m]
        Length of nose
    D_bearing1 : float, [m]
        Diameter of bearing #1 (closer to hub)
    D_bearing2 : float, [m]
        Diameter of bearing #2 (closer to tower)
    s_nose : numpy array[5], [m]
        Nose discretized hub-aligned s-coordinates
    nose_mass : float, [kg]
        Nose mass
    nose_cm : float, [m]
        Nose center of mass along nose axis from bedplate
    nose_I : numpy array[3], [kg*m**2]
        Nose moment of inertia around cm in axial (hub-aligned) c.s.
    x_bedplate : numpy array[12], [m]
        Bedplate centerline x-coordinates
    z_bedplate : numpy array[12], [m]
        Bedplate centerline z-coordinates
    x_bedplate_inner : numpy array[12], [m]
        Bedplate lower curve x-coordinates
    z_bedplate_inner : numpy array[12], [m]
        Bedplate lower curve z-coordinates
    x_bedplate_outer : numpy array[12], [m]
        Bedplate outer curve x-coordinates
    z_bedplate_outer : numpy array[12], [m]
        Bedplate outer curve z-coordinates
    D_bedplate : numpy array[12], [m]
        Bedplate diameters
    t_bedplate : numpy array[12], [m]
        Bedplate wall thickness (mirrors input)
    s_stator : float, [m]
        Generator stator attachment to nose s-coordinate
    s_rotor : float, [m]
        Generator rotor attachment to lss s-coordinate
    constr_access : numpy array[2], [m]
        Margin for allowing maintenance access (should be > 0)
    constr_ecc : float, [m]
        Margin for bedplate ellipse eccentricity (should be > 0)
    """

    def setup(self):
        super().setup()

        self.add_input("access_diameter", 0.0, units="m")
        self.add_input("nose_diameter", np.zeros(2), units="m")
        self.add_input("nose_wall_thickness", np.zeros(2), units="m")
        self.add_input("bedplate_wall_thickness", np.zeros(4), units="m")

        self.add_output("L_nose", 0.0, units="m")
        self.add_output("D_bearing1", 0.0, units="m")
        self.add_output("D_bearing2", 0.0, units="m")
        self.add_output("s_nose", val=np.zeros(5), units="m")
        self.add_output("nose_mass", val=0.0, units="kg")
        self.add_output("nose_cm", val=0.0, units="m")
        self.add_output("nose_I", val=np.zeros(3), units="kg*m**2")
        self.add_output("x_bedplate", val=np.zeros(12), units="m")
        self.add_output("z_bedplate", val=np.zeros(12), units="m")
        self.add_output("x_bedplate_inner", val=np.zeros(12), units="m")
        self.add_output("z_bedplate_inner", val=np.zeros(12), units="m")
        self.add_output("x_bedplate_outer", val=np.zeros(12), units="m")
        self.add_output("z_bedplate_outer", val=np.zeros(12), units="m")
        self.add_output("D_bedplate", val=np.zeros(12), units="m")
        self.add_output("t_bedplate", val=np.zeros(12), units="m")
        self.add_output("constr_access", np.zeros((2, 2)), units="m")
        self.add_output("constr_ecc", 0.0, units="m")

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        # Unpack inputs
        L_12 = float(inputs["L_12"][0])
        L_h1 = float(inputs["L_h1"][0])
        L_generator = float(inputs["L_generator"][0])
        L_overhang = float(inputs["overhang"][0])
        H_drive = float(inputs["drive_height"][0])
        tilt = float(np.deg2rad(inputs["tilt"][0]))
        D_access = float(inputs["access_diameter"][0])
        D_nose = inputs["nose_diameter"]
        D_lss = inputs["lss_diameter"]
        D_top = float(inputs["D_top"][0])
        D_hub = float(inputs["hub_diameter"][0])
        t_nose = inputs["nose_wall_thickness"]
        t_lss = inputs["lss_wall_thickness"]
        t_bed = inputs["bedplate_wall_thickness"]
        upwind = discrete_inputs["upwind"]
        lss_rho = float(inputs["lss_rho"][0])
        bedplate_rho = float(inputs["bedplate_rho"][0])
        bedplate_mass_user = float(inputs["bedplate_mass_user"][0])

        # ------- Discretization ----------------
        L_grs = 0.5 * L_h1
        L_gsn = L_generator - L_grs - L_12
        L_2n = 2.0 * L_gsn

        # Length of lss and nose
        L_lss = L_12 + L_h1
        L_nose = L_12 + L_2n
        outputs["L_lss"] = L_lss
        outputs["L_nose"] = L_nose

        # Total length from bedplate to hub flange
        ds = 0.5 * np.ones(2)
        s_drive = np.cumsum(np.r_[0.0, L_2n * ds, L_12 * ds, L_h1 * ds])
        L_drive = s_drive[-1]
        outputs["L_drive"] = L_drive

        # From Overhang input (dist from center of tower measured in yaw-aligned
        # c.s.-parallel to ground), compute bedplate length and height
        L_bedplate = L_overhang - (L_drive + 0.5 * D_hub) * np.cos(tilt)
        constr_Ldrive = L_bedplate - 0.5 * D_top  # Should be > 0
        if constr_Ldrive < 0:
            L_bedplate = 0.5 * D_top
        H_bedplate = H_drive - (L_drive + 0.5 * D_hub) * np.sin(tilt)  # Keep eccentricity under control
        outputs["L_bedplate"] = L_bedplate
        outputs["H_bedplate"] = H_bedplate

        # Discretize the drivetrain from bedplate end to hub
        # Note that the offset from the bedplate to the tower center line is taken into account in drive_components by "shaft0"
        s_mb1 = s_drive[4]
        s_mb2 = s_drive[2]
        s_rotor = s_drive[-2]
        s_stator = s_drive[1]
        s_nose = s_drive[:5]
        s_lss = s_drive[2:]

        # Store outputs
        # outputs['s_drive']     = np.sort(s_drive)
        outputs["s_rotor"] = s_rotor
        outputs["s_stator"] = s_stator
        outputs["s_nose"] = s_nose
        outputs["s_lss"] = s_lss
        outputs["s_generator"] = 0.5 * (s_rotor + s_stator)
        outputs["s_mb1"] = s_mb1
        outputs["s_mb2"] = s_mb2
        # ------------------------------------

        # ------------ Bedplate geometry and coordinates -------------
        # Define reference/centroidal axis
        # Origin currently set like standard ellipse eqns, but will shift below to being at tower top
        # The end point of 90 deg isn't exactly right for non-zero tilt, but leaving that for later
        n_points = 12
        if upwind:
            rad = np.linspace(0.0, 0.5 * np.pi, n_points)
            dw = 1
        else:
            rad = np.linspace(np.pi, 0.5 * np.pi, n_points)
            dw = -1

        # Make sure we have the right number of bedplate thickness points
        t_bed = np.interp(rad, np.linspace(rad[0], rad[-1], len(t_bed)), t_bed)

        # Centerline
        x_c = L_bedplate * np.cos(rad)
        z_c = H_bedplate * np.sin(rad)

        # Points on the outermost ellipse
        x_outer = (L_bedplate + 0.5 * D_top) * np.cos(rad)
        z_outer = (H_bedplate + 0.5 * D_nose[0]) * np.sin(rad)

        # Points on the innermost ellipse
        x_inner = (L_bedplate - 0.5 * D_top) * np.cos(rad)
        z_inner = (H_bedplate - 0.5 * D_nose[0]) * np.sin(rad)

        # Cross-sectional properties
        D_bed = np.sqrt((z_outer - z_inner) ** 2 + (x_outer - x_inner) ** 2)
        r_bed_o = 0.5 * D_bed
        r_bed_i = r_bed_o - t_bed
        A_bed = np.pi * (r_bed_o**2 - r_bed_i**2)

        # This finds the central angle (rad2) given the parametric angle (rad)
        rad2 = np.arctan2(z_c/H_bedplate, dw * x_c/L_bedplate)

        # arc length from eccentricity of the centroidal ellipse using incomplete elliptic integral of the second kind
        if L_bedplate >= H_bedplate:
            ecc = np.sqrt(1 - (H_bedplate / L_bedplate) ** 2)
            arc = L_bedplate * np.diff(ellipeinc(rad2, ecc))
        else:
            ecc = np.sqrt(1 - (L_bedplate / H_bedplate) ** 2)
            arc = H_bedplate * np.diff(ellipeinc(rad2, ecc))

        # Mass and MoI properties
        x_c_sec = util.nodal2sectional(x_c)[0]
        z_c_sec = util.nodal2sectional(z_c)[0]
        # R_c_sec = np.sqrt( x_c_sec**2 + z_c_sec**2 ) # unnecesary
        mass = util.nodal2sectional(A_bed)[0] * arc * bedplate_rho
        mass_tot = mass.sum()
        cm = np.array([np.sum(mass * x_c_sec), 0.0, np.sum(mass * z_c_sec)]) / mass_tot
        # For I, could do integral over sectional I, rotate axes by rad2, and then parallel axis theorem
        # we simplify by assuming lumped point mass.  TODO: Find a good way to check this?  Torus shell?
        I_bed = util.assembleI(np.zeros(6))
        for k in range(len(mass)):
            r_bed_o_k = 0.5 * (r_bed_o[k] + r_bed_o[k + 1])
            r_bed_i_k = 0.5 * (r_bed_i[k] + r_bed_i[k + 1])
            I_sec = mass[k] * np.array(
                [
                    0.5 * (r_bed_o_k**2 + r_bed_i_k**2),
                    (1.0 / 12.0) * (3 * (r_bed_o_k**2 + r_bed_i_k**2) + arc[k] ** 2),
                    (1.0 / 12.0) * (3 * (r_bed_o_k**2 + r_bed_i_k**2) + arc[k] ** 2),
                ]
            )
            I_sec_rot = util.rotateI(I_sec, 0.5 * np.pi - rad2[k], axis="y")
            R_k = np.array([x_c_sec[k] - x_c[0], 0.0, z_c_sec[k]])
            I_bed += util.assembleI(I_sec_rot) + mass[k] * (np.dot(R_k, R_k) * np.eye(3) - np.outer(R_k, R_k))

        # Now shift origin to be at tower top
        cm[0] -= x_c[0]
        x_inner -= x_c[0]
        x_outer -= x_c[0]
        x_c -= x_c[0]

        coef_user = 1.0 if bedplate_mass_user == 0.0 else bedplate_mass_user / mass_tot
        outputs["bedplate_mass"] = coef_user * mass_tot
        outputs["bedplate_cm"] = cm
        outputs["bedplate_I"] = coef_user * util.unassembleI(I_bed)

        # Geometry outputs
        outputs["x_bedplate"] = x_c
        outputs["z_bedplate"] = z_c
        outputs["x_bedplate_inner"] = x_inner
        outputs["z_bedplate_inner"] = z_inner
        outputs["x_bedplate_outer"] = x_outer
        outputs["z_bedplate_outer"] = z_outer
        outputs["D_bedplate"] = D_bed
        outputs["t_bedplate"] = t_bed
        # ------------------------------------

        # ------- Constraints ----------------
        outputs["constr_access"] = np.c_[D_lss - 2 * t_lss - D_nose - 0.25 * D_access, D_nose - 2 * t_nose - D_access]
        outputs["constr_length"] = constr_Ldrive  # Should be > 0
        outputs["constr_height"] = H_bedplate  # Should be > 0
        outputs["constr_ecc"] = L_bedplate - H_bedplate  # Should be > 0
        # ------------------------------------

        # ------- Nose, lss, and bearing properties ----------------
        # Now is a good time to set bearing diameters
        outputs["D_bearing1"] = 0.5 * D_lss[-1] - t_lss[-1] - 0.5 * D_nose[0]
        outputs["D_bearing2"] = 0.5 * D_lss[-1] - t_lss[-1] - 0.5 * D_nose[-1]

        # Compute center of mass based on area
        m_nose, cm_nose, I_nose = rod_prop(s_nose, D_nose, t_nose, bedplate_rho)
        outputs["nose_mass"] = m_nose
        outputs["nose_cm"] = cm_nose
        outputs["nose_I"] = I_nose

        m_lss, cm_lss, I_lss = rod_prop(s_lss, D_lss, t_lss, lss_rho)
        outputs["lss_mass"] = m_lss
        outputs["lss_cm"] = cm_lss
        outputs["lss_I"] = I_lss


class GearedLayout(Layout):
    """
    Calculate lengths, heights, and diameters of key drivetrain components in a
    geared drive system (valid for upwind or downwind).

    |_Lgen|_Lhss|Lgear|dl|_L12_|_Lh1_|
                      |_____Llss_____|
    |--|--|--|--|--|--|--|--|--|--|--|
    0  1  2  3  4  5  6  7  8  9  10 11 (indices)
                        mb2   mb1

    Parameters
    ----------
    hss_diameter : numpy array[2], [m]
        HSS outer diameter from hub to bearing 2
    hss_wall_thickness : numpy array[2], [m]
        HSS wall thickness
    bedplate_flange_width : float, [m]
        Bedplate is two parallel I beams, this is the flange width
    bedplate_flange_thickness : float, [m]
        Bedplate is two parallel I beams, this is the flange thickness
    bedplate_web_thickness : float, [m]
        Bedplate is two parallel I beams, this is the web thickness
    bedplate_web_height : float, [m]
        Bedplate is two parallel I beams, this is the web height
    hss_rho : float, [kg/m**3]
        material density

    Returns
    -------
    s_drive : numpy array[12], [m]
        Discretized, hub-aligned s-coordinates of the drivetrain starting at
        generator and ending at hub flange
    s_hss : numpy array[5], [m]
        HSS discretized s-coordinates
    hss_mass : float, [kg]
        HSS mass
    hss_cm : float, [m]
        HSS center of mass along hss axis from bedplate
    hss_I : numpy array[3], [kg*m**2]
        HSS moment of inertia around cm in axial (hub-aligned) c.s.
    s_gearbox : float, [m]
        Gearbox (centroid) position in s-coordinates
    s_generator : float, [m]
        Generator (centroid) position in s-coordinates

    """

    def setup(self):
        super().setup()

        self.add_input("L_hss", 0.0, units="m")
        self.add_input("L_gearbox", 0.0, units="m")
        self.add_input("hss_diameter", np.zeros(2), units="m")
        self.add_input("hss_wall_thickness", np.zeros(2), units="m")
        self.add_input("hss_rho", val=0.0, units="kg/m**3")
        self.add_input("bedplate_flange_width", val=0.0, units="m")
        self.add_input("bedplate_flange_thickness", val=0.0, units="m")
        self.add_input("bedplate_web_thickness", val=0.0, units="m")

        self.add_output("s_drive", val=np.zeros(12), units="m")
        self.add_output("s_hss", val=np.zeros(3), units="m")
        self.add_output("bedplate_web_height", val=0.0, units="m")

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        # Unpack inputs
        upwind = discrete_inputs["upwind"]
        Cup = -1.0 if upwind else 1.0

        L_12 = float(inputs["L_12"][0])
        L_h1 = float(inputs["L_h1"][0])
        L_hss = float(inputs["L_hss"][0])
        L_gearbox = float(inputs["L_gearbox"][0])
        L_generator = float(inputs["L_generator"][0])
        L_overhang = float(inputs["overhang"][0])
        H_drive = float(inputs["drive_height"][0])

        tilt = float(np.deg2rad(inputs["tilt"][0]))

        D_lss = inputs["lss_diameter"]
        t_lss = inputs["lss_wall_thickness"]
        D_hss = inputs["hss_diameter"]
        t_hss = inputs["hss_wall_thickness"]

        D_top = float(inputs["D_top"][0])
        D_hub = float(inputs["hub_diameter"][0])

        bed_w_flange = float(inputs["bedplate_flange_width"][0])
        bed_t_flange = float(inputs["bedplate_flange_thickness"][0])
        # bed_h_web    = float(inputs['bedplate_web_height'][0])
        bed_t_web = float(inputs["bedplate_web_thickness"][0])

        lss_rho = float(inputs["lss_rho"][0])
        hss_rho = float(inputs["hss_rho"][0])
        bedplate_rho = float(inputs["bedplate_rho"][0])
        bedplate_mass_user = float(inputs["bedplate_mass_user"][0])

        # ------- Discretization ----------------
        # Length of lss and drivetrain length
        delta = 0.1  # separation between MB2 and gearbox attachment
        L_lss = L_12 + L_h1 + delta
        L_drive = L_lss + L_gearbox + L_hss + L_generator
        ds = 0.5 * np.ones(2)
        s_drive = np.cumsum(np.r_[0.0, L_generator * ds, L_hss * ds, L_gearbox * ds, delta, L_12 * ds, L_h1 * ds])
        L_drive = s_drive[-1] - s_drive[0]
        outputs["L_drive"] = L_drive
        outputs["L_lss"] = L_lss

        # Put tower at 0 position
        s_tower = s_drive[-1] + 0.5 * D_hub - L_overhang / np.cos(tilt)
        s_drive -= s_tower
        outputs["s_drive"] = s_drive

        # Discretize the drivetrain from generator to hub
        s_stator = s_drive[0]
        s_rotor = s_drive[1]
        s_generator = 0.5*(s_rotor + s_stator)
        s_mb1 = s_drive[9]
        s_mb2 = s_drive[7]
        s_gearbox = s_drive[5]
        s_lss = s_drive[6:]
        s_lss = np.r_[s_lss[:-2], s_lss[-1]]  # Need to stick to 5 points
        s_hss = s_drive[2:5]

        # Store outputs
        outputs["s_stator"] = s_stator
        outputs["s_rotor"] = s_rotor
        outputs["s_generator"] = s_generator
        outputs["s_gearbox"] = s_gearbox
        outputs["s_mb1"] = s_mb1
        outputs["s_mb2"] = s_mb2
        # ------------------------------------

        # ------- hss, lss, and bearing properties ----------------
        # Compute center of mass based on area
        m_hss, cm_hss, I_hss = rod_prop(s_hss, D_hss, t_hss, hss_rho)
        outputs["hss_mass"] = m_hss
        outputs["hss_cm"] = cm_hss
        outputs["hss_I"] = I_hss
        outputs["s_hss"] = s_hss

        m_lss, cm_lss, I_lss = rod_prop(s_lss, D_lss, t_lss, lss_rho)
        outputs["lss_mass"] = m_lss
        outputs["lss_cm"] = cm_lss
        outputs["lss_I"] = I_lss
        outputs["s_lss"] = s_lss

        # ------- Bedplate I-beam properties ----------------
        L_bedplate = (L_drive + 0.5 * D_hub) * np.cos(tilt)
        H_bedplate = H_drive - (L_drive + 0.5 * D_hub) * np.sin(tilt)  # Subtract thickness of platform plate
        outputs["L_bedplate"] = L_bedplate
        outputs["H_bedplate"] = H_bedplate
        bed_h_web = H_bedplate - 2 * bed_t_flange - 0.05  # Leave some extra room for plate?

        yoff = 0.25 * D_top
        myI = IBeam(bed_w_flange, bed_t_flange, bed_h_web, bed_t_web)
        m_bedplate = myI.Area * L_bedplate * bedplate_rho
        cg_bedplate = np.r_[Cup * (L_overhang - 0.5 * L_bedplate), 0.0, myI.CG]  # from tower top
        I_bedplate = (
            bedplate_rho * L_bedplate * np.r_[myI.J0, myI.Ixx, myI.Iyy]
            + m_bedplate * L_bedplate**2 / 12.0 * np.r_[0.0, 1.0, 1.0]
            + m_bedplate * yoff**2 * np.r_[1.0, 0.0, 1.0]
        )
        coef_user = 1.0 if bedplate_mass_user == 0.0 else bedplate_mass_user / (2 * m_bedplate)
        outputs["bedplate_web_height"] = bed_h_web
        outputs["bedplate_mass"] = coef_user * 2 * m_bedplate
        outputs["bedplate_cm"] = cg_bedplate
        outputs["bedplate_I"] = coef_user * 2 * np.r_[I_bedplate, np.zeros(3)]

        # ------- Constraints ----------------
        outputs["constr_length"] = (L_drive + 0.5 * D_hub) * np.cos(tilt) - L_overhang - 0.5 * D_top  # Should be > 0
        outputs["constr_height"] = H_bedplate  # Should be > 0
        # ------------------------------------
